/* Generated by Cython 0.24.0a0 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24_0a0"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__s3ql__deltadump
#define __PYX_HAVE_API__s3ql__deltadump
#include "string.h"
#include "stdio.h"
#include "errno.h"
#include "stdlib.h"
#include "stdint.h"
#include "sys/types.h"
#include "unistd.h"
#include "endian_indep.h"
#include "sqlite3.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/s3ql/deltadump.pyx",
  "type.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table;
struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table;
struct __pyx_opt_args_4s3ql_9deltadump_raise_from_errno;

/* "s3ql/deltadump.pyx":136
 *     return None
 * 
 * cdef int raise_from_errno(err_class=OSError) except -1:             # <<<<<<<<<<<<<<
 *     '''Raise OSError for current errno value'''
 * 
 */
struct __pyx_opt_args_4s3ql_9deltadump_raise_from_errno {
  int __pyx_n;
  PyObject *err_class;
};

/* "s3ql/deltadump.pyx":280
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 * def dump_table(table, order, columns, db, fh):             # <<<<<<<<<<<<<<
 *     '''Dump *columns* of *table* into *fh*
 * 
 */
struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table {
  PyObject_HEAD
  int *__pyx_v_col_args;
  int *__pyx_v_col_types;
  FILE *__pyx_v_fp;
  int64_t *__pyx_v_int64_prev;
  sqlite3 *__pyx_v_sqlite3_db;
  sqlite3_stmt *__pyx_v_stmt;
};


/* "s3ql/deltadump.pyx":401
 *                         fwrite(buf, len_, fp)
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):             # <<<<<<<<<<<<<<
 *     '''Load *columns* of *table* from *fh*
 * 
 */
struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table {
  PyObject_HEAD
  sqlite3_stmt *__pyx_v_begin_stmt;
  void *__pyx_v_buf;
  int *__pyx_v_col_args;
  int *__pyx_v_col_types;
  sqlite3_stmt *__pyx_v_commit_stmt;
  FILE *__pyx_v_fp;
  int64_t *__pyx_v_int64_prev;
  sqlite3 *__pyx_v_sqlite3_db;
  sqlite3_stmt *__pyx_v_stmt;
};


/* --- Runtime support code (head) --- */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

#include <string.h>

static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_EqObjC(op1, op2, intval, inplace)\
    PyObject_RichCompare(op1, op2, Py_EQ)
    #endif

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

typedef struct {
    int code_line;
    PyCodeObject* code_object;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

static int __Pyx_check_binary_version(void);

#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

static PyObject *__Pyx_ImportModule(const char *name);

static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'cpython.long' */

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'libc.errno' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'libc.stdint' */

/* Module declarations from 'posix.types' */

/* Module declarations from 'posix.unistd' */

/* Module declarations from 's3ql.deltadump' */
static PyTypeObject *__pyx_ptype_4s3ql_9deltadump___pyx_scope_struct__dump_table = 0;
static PyTypeObject *__pyx_ptype_4s3ql_9deltadump___pyx_scope_struct_1_load_table = 0;
static int __pyx_v_4s3ql_9deltadump__INTEGER;
static int __pyx_v_4s3ql_9deltadump__BLOB;
static int __pyx_v_4s3ql_9deltadump__TIME;
static uint8_t __pyx_v_4s3ql_9deltadump_INT8;
static uint8_t __pyx_v_4s3ql_9deltadump_INT16;
static uint8_t __pyx_v_4s3ql_9deltadump_INT32;
static uint8_t __pyx_v_4s3ql_9deltadump_INT64;
static double __pyx_v_4s3ql_9deltadump_time_scale;
static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_fwrite(const void *, size_t, FILE *); /*proto*/
static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_fread(void *, size_t, FILE *); /*proto*/
static PyObject *__pyx_f_4s3ql_9deltadump_free(void *); /*proto*/
static int __pyx_f_4s3ql_9deltadump_raise_from_errno(struct __pyx_opt_args_4s3ql_9deltadump_raise_from_errno *__pyx_optional_args); /*proto*/
static int __pyx_f_4s3ql_9deltadump_fclose(FILE *); /*proto*/
static void *__pyx_f_4s3ql_9deltadump_calloc(size_t, size_t); /*proto*/
static int __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(int, int, sqlite3 *); /*proto*/
static int __pyx_f_4s3ql_9deltadump_prep_columns(PyObject *, int **, int **); /*proto*/
static FILE *__pyx_f_4s3ql_9deltadump_dup_to_fp(PyObject *, const char *); /*proto*/
static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_write_integer(int64_t, FILE *); /*proto*/
static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_read_integer(int64_t *, FILE *); /*proto*/
#define __Pyx_MODULE_NAME "s3ql.deltadump"
int __pyx_module_is_main_s3ql__deltadump = 0;

/* Implementation of 's3ql.deltadump' */
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_IOError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_RuntimeError;
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k__4[] = ", ";
static const char __pyx_k__9[] = "?";
static const char __pyx_k_cm[] = "cm";
static const char __pyx_k_db[] = "db";
static const char __pyx_k_fh[] = "fh";
static const char __pyx_k_fp[] = "fp";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_rc[] = "rc";
static const char __pyx_k_buf[] = "buf";
static const char __pyx_k_cmd[] = "cmd";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_len[] = "len_";
static const char __pyx_k_log[] = "log";
static const char __pyx_k_sys[] = "sys";
static const char __pyx_k_tmp[] = "tmp";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_BLOB[] = "BLOB";
static const char __pyx_k_TIME[] = "TIME";
static const char __pyx_k_apsw[] = "apsw";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_stmt[] = "stmt";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_int64[] = "int64";
static const char __pyx_k_order[] = "order";
static const char __pyx_k_query[] = "query";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_table[] = "table";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_fileno[] = "fileno";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_memory[] = ":memory:";
static const char __pyx_k_pragma[] = "pragma";
static const char __pyx_k_INTEGER[] = "INTEGER";
static const char __pyx_k_IOError[] = "IOError";
static const char __pyx_k_OSError[] = "OSError";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_get_val[] = "get_val";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_PRAGMA_s[] = "PRAGMA %s";
static const char __pyx_k_callback[] = "callback";
static const char __pyx_k_col_args[] = "col_args";
static const char __pyx_k_dbfile_b[] = "dbfile_b";
static const char __pyx_k_trx_rows[] = "trx_rows";
static const char __pyx_k_ExitStack[] = "ExitStack";
static const char __pyx_k_col_count[] = "col_count";
static const char __pyx_k_col_names[] = "col_names";
static const char __pyx_k_col_types[] = "col_types";
static const char __pyx_k_getLogger[] = "getLogger";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_row_count[] = "row_count";
static const char __pyx_k_PRAGMA_s_s[] = "PRAGMA %s = %s";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_begin_stmt[] = "begin_stmt";
static const char __pyx_k_contextlib[] = "contextlib";
static const char __pyx_k_dump_table[] = "dump_table";
static const char __pyx_k_int64_prev[] = "int64_prev";
static const char __pyx_k_load_table[] = "load_table";
static const char __pyx_k_sqlite3_db[] = "sqlite3_db";
static const char __pyx_k_commit_stmt[] = "commit_stmt";
static const char __pyx_k_synchronous[] = "synchronous";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_check_sqlite[] = "check_sqlite";
static const char __pyx_k_exceptionfor[] = "exceptionfor";
static const char __pyx_k_foreign_keys[] = "foreign_keys";
static const char __pyx_k_MAX_BLOB_SIZE[] = "MAX_BLOB_SIZE";
static const char __pyx_k_s3ql_deltadump[] = "s3ql.deltadump";
static const char __pyx_k_Corrupted_input[] = "Corrupted input";
static const char __pyx_k_compile_options[] = "compile_options";
static const char __pyx_k_surrogateescape[] = "surrogateescape";
static const char __pyx_k_sqlitelibversion[] = "sqlitelibversion";
static const char __pyx_k_BEGIN_TRANSACTION[] = "BEGIN TRANSACTION";
static const char __pyx_k_COMMIT_TRANSACTION[] = "COMMIT TRANSACTION";
static const char __pyx_k_apsw_sqlite_options[] = "apsw_sqlite_options";
static const char __pyx_k_apsw_sqlite_version[] = "apsw_sqlite_version";
static const char __pyx_k_s3ql_sqlite_options[] = "s3ql_sqlite_options";
static const char __pyx_k_s3ql_sqlite_version[] = "s3ql_sqlite_version";
static const char __pyx_k_getfilesystemencoding[] = "getfilesystemencoding";
static const char __pyx_k_Can_t_dump_NULL_values[] = "Can't dump NULL values";
static const char __pyx_k_Length_d_d_in_column_d[] = "Length %d != %d in column %d";
static const char __pyx_k_Opening_connection_to_s[] = "Opening connection to %s";
static const char __pyx_k_INSERT_INTO_s_s_VALUES_s[] = "INSERT INTO %s (%s) VALUES(%s)";
static const char __pyx_k_dump_table_locals_lambda[] = "dump_table.<locals>.<lambda>";
static const char __pyx_k_load_table_locals_lambda[] = "load_table.<locals>.<lambda>";
static const char __pyx_k_Invalid_type_for_column_d[] = "Invalid type for column %d";
static const char __pyx_k_SELECT_COUNT_rowid_FROM_s[] = "SELECT COUNT(rowid) FROM %s";
static const char __pyx_k_SELECT_s_FROM_s_ORDER_BY_s[] = "SELECT %s FROM %s ORDER BY %s ";
static const char __pyx_k_dump_table_s_writing_d_rows[] = "dump_table(%s): writing %d rows";
static const char __pyx_k_load_table_s_reading_d_rows[] = "load_table(%s): reading %d rows";
static const char __pyx_k_BLOB_too_large_to_read_d_vs_d[] = "BLOB too large to read (%d vs %d)";
static const char __pyx_k_home_nikratio_in_progress_s3ql[] = "/home/nikratio/in-progress/s3ql/src/s3ql/deltadump.pyx";
static const char __pyx_k_Can_not_dump_BLOB_of_size_d_max[] = "Can not dump BLOB of size %d (max: %d)";
static const char __pyx_k_SQLite_version_mismatch_between[] = "SQLite version mismatch between APSW and S3QL (%s vs %s)";
static const char __pyx_k_deltadump_pyx_this_file_is_part[] = "\ndeltadump.pyx - this file is part of S3QL (http://s3ql.googlecode.com)\n\nCopyright \302\251 2008 Nikolaus Rath <Nikolaus@rath.org>\n\nThis program can be distributed under the terms of the GNU GPLv3.\n";
static const char __pyx_k_Can_t_access_in_memory_databases[] = "Can't access in-memory databases";
static const char __pyx_k_SQLite_code_used_by_APSW_was_com[] = "SQLite code used by APSW was compiled with different options than SQLite code available to S3QL! Differing settings: + %s, - %s";
static const char __pyx_k_apsw_sqlite_compile_options_s_s3[] = "apsw sqlite compile options: %s, s3ql sqlite compile options: %s";
static const char __pyx_k_apsw_sqlite_version_s_s3ql_sqlit[] = "apsw sqlite version: %s, s3ql sqlite version: %s";
static PyObject *__pyx_kp_b_BEGIN_TRANSACTION;
static PyObject *__pyx_n_s_BLOB;
static PyObject *__pyx_kp_u_BLOB_too_large_to_read_d_vs_d;
static PyObject *__pyx_kp_b_COMMIT_TRANSACTION;
static PyObject *__pyx_kp_u_Can_not_dump_BLOB_of_size_d_max;
static PyObject *__pyx_kp_u_Can_t_access_in_memory_databases;
static PyObject *__pyx_kp_u_Can_t_dump_NULL_values;
static PyObject *__pyx_kp_u_Corrupted_input;
static PyObject *__pyx_n_s_ExitStack;
static PyObject *__pyx_kp_u_INSERT_INTO_s_s_VALUES_s;
static PyObject *__pyx_n_s_INTEGER;
static PyObject *__pyx_n_s_IOError;
static PyObject *__pyx_kp_u_Invalid_type_for_column_d;
static PyObject *__pyx_kp_u_Length_d_d_in_column_d;
static PyObject *__pyx_n_s_MAX_BLOB_SIZE;
static PyObject *__pyx_n_s_OSError;
static PyObject *__pyx_kp_u_Opening_connection_to_s;
static PyObject *__pyx_kp_u_PRAGMA_s;
static PyObject *__pyx_kp_u_PRAGMA_s_s;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_kp_u_SELECT_COUNT_rowid_FROM_s;
static PyObject *__pyx_kp_u_SELECT_s_FROM_s_ORDER_BY_s;
static PyObject *__pyx_kp_u_SQLite_code_used_by_APSW_was_com;
static PyObject *__pyx_kp_u_SQLite_version_mismatch_between;
static PyObject *__pyx_n_s_TIME;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_u__4;
static PyObject *__pyx_kp_u__9;
static PyObject *__pyx_n_s_apsw;
static PyObject *__pyx_kp_u_apsw_sqlite_compile_options_s_s3;
static PyObject *__pyx_n_s_apsw_sqlite_options;
static PyObject *__pyx_n_s_apsw_sqlite_version;
static PyObject *__pyx_kp_u_apsw_sqlite_version_s_s3ql_sqlit;
static PyObject *__pyx_n_s_begin_stmt;
static PyObject *__pyx_n_s_buf;
static PyObject *__pyx_n_s_callback;
static PyObject *__pyx_n_s_check_sqlite;
static PyObject *__pyx_n_s_cm;
static PyObject *__pyx_n_s_cmd;
static PyObject *__pyx_n_s_col_args;
static PyObject *__pyx_n_s_col_count;
static PyObject *__pyx_n_s_col_names;
static PyObject *__pyx_n_s_col_types;
static PyObject *__pyx_n_s_columns;
static PyObject *__pyx_n_s_commit_stmt;
static PyObject *__pyx_n_s_compile_options;
static PyObject *__pyx_n_s_contextlib;
static PyObject *__pyx_n_s_count;
static PyObject *__pyx_n_s_db;
static PyObject *__pyx_n_s_dbfile_b;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_dump_table;
static PyObject *__pyx_n_s_dump_table_locals_lambda;
static PyObject *__pyx_kp_u_dump_table_s_writing_d_rows;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_exceptionfor;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_fh;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_fileno;
static PyObject *__pyx_n_u_foreign_keys;
static PyObject *__pyx_n_s_fp;
static PyObject *__pyx_n_s_getLogger;
static PyObject *__pyx_n_s_get_val;
static PyObject *__pyx_n_s_getfilesystemencoding;
static PyObject *__pyx_kp_s_home_nikratio_in_progress_s3ql;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_idx;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_int64;
static PyObject *__pyx_n_s_int64_prev;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_load_table;
static PyObject *__pyx_n_s_load_table_locals_lambda;
static PyObject *__pyx_kp_u_load_table_s_reading_d_rows;
static PyObject *__pyx_n_s_log;
static PyObject *__pyx_n_s_logging;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_kp_u_memory;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_order;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_pragma;
static PyObject *__pyx_n_s_query;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rc;
static PyObject *__pyx_n_s_row_count;
static PyObject *__pyx_n_s_s3ql_deltadump;
static PyObject *__pyx_n_s_s3ql_sqlite_options;
static PyObject *__pyx_n_s_s3ql_sqlite_version;
static PyObject *__pyx_n_s_sqlite3_db;
static PyObject *__pyx_n_s_sqlitelibversion;
static PyObject *__pyx_n_s_stmt;
static PyObject *__pyx_n_u_surrogateescape;
static PyObject *__pyx_n_u_synchronous;
static PyObject *__pyx_n_s_sys;
static PyObject *__pyx_n_s_table;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tmp;
static PyObject *__pyx_n_s_trx_rows;
static PyObject *__pyx_n_s_val;
static PyObject *__pyx_pf_4s3ql_9deltadump_check_sqlite(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda1(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda2(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda3(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda4(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda5(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4s3ql_9deltadump_2dump_table(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_table, PyObject *__pyx_v_order, PyObject *__pyx_v_columns, PyObject *__pyx_v_db, PyObject *__pyx_v_fh); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda6(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda7(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda8(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda9(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda10(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda11(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda12(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda13(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda14(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda15(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4s3ql_9deltadump_4load_table(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_table, PyObject *__pyx_v_columns, PyObject *__pyx_v_db, PyObject *__pyx_v_fh, PyObject *__pyx_v_trx_rows); /* proto */
static PyObject *__pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct__dump_table(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct_1_load_table(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_4096;
static PyObject *__pyx_int_5000;
static PyObject *__pyx_k_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__17;

/* "s3ql/deltadump.pyx":112
 * cdef double time_scale = 1 << 30
 * 
 * cdef inline int fwrite(const_void * buf, size_t len_, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Call libc's fwrite() and raise exception on failure'''
 * 
 */

static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_fwrite(const void *__pyx_v_buf, size_t __pyx_v_len_, FILE *__pyx_v_fp) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_opt_args_4s3ql_9deltadump_raise_from_errno __pyx_t_3;
  __Pyx_RefNannySetupContext("fwrite", 0);

  /* "s3ql/deltadump.pyx":115
 *     '''Call libc's fwrite() and raise exception on failure'''
 * 
 *     if fwrite_c(buf, len_, 1, fp) != 1:             # <<<<<<<<<<<<<<
 *         raise_from_errno(IOError)
 *     return 0
 */
  __pyx_t_1 = ((fwrite(__pyx_v_buf, __pyx_v_len_, 1, __pyx_v_fp) != 1) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":116
 * 
 *     if fwrite_c(buf, len_, 1, fp) != 1:
 *         raise_from_errno(IOError)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_t_3.__pyx_n = 1;
    __pyx_t_3.err_class = __pyx_builtin_IOError;
    __pyx_t_2 = __pyx_f_4s3ql_9deltadump_raise_from_errno(&__pyx_t_3); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 116, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":115
 *     '''Call libc's fwrite() and raise exception on failure'''
 * 
 *     if fwrite_c(buf, len_, 1, fp) != 1:             # <<<<<<<<<<<<<<
 *         raise_from_errno(IOError)
 *     return 0
 */
  }

  /* "s3ql/deltadump.pyx":117
 *     if fwrite_c(buf, len_, 1, fp) != 1:
 *         raise_from_errno(IOError)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef inline int fread(void * buf, size_t len_, FILE * fp) except -1:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":112
 * cdef double time_scale = 1 << 30
 * 
 * cdef inline int fwrite(const_void * buf, size_t len_, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Call libc's fwrite() and raise exception on failure'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("s3ql.deltadump.fwrite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":119
 *     return 0
 * 
 * cdef inline int fread(void * buf, size_t len_, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Call libc's fread() and raise exception on failure'''
 * 
 */

static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_fread(void *__pyx_v_buf, size_t __pyx_v_len_, FILE *__pyx_v_fp) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  struct __pyx_opt_args_4s3ql_9deltadump_raise_from_errno __pyx_t_3;
  __Pyx_RefNannySetupContext("fread", 0);

  /* "s3ql/deltadump.pyx":122
 *     '''Call libc's fread() and raise exception on failure'''
 * 
 *     if fread_c(buf, len_, 1, fp) != 1:             # <<<<<<<<<<<<<<
 *         raise_from_errno(IOError)
 *     return 0
 */
  __pyx_t_1 = ((fread(__pyx_v_buf, __pyx_v_len_, 1, __pyx_v_fp) != 1) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":123
 * 
 *     if fread_c(buf, len_, 1, fp) != 1:
 *         raise_from_errno(IOError)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_t_3.__pyx_n = 1;
    __pyx_t_3.err_class = __pyx_builtin_IOError;
    __pyx_t_2 = __pyx_f_4s3ql_9deltadump_raise_from_errno(&__pyx_t_3); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 123, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":122
 *     '''Call libc's fread() and raise exception on failure'''
 * 
 *     if fread_c(buf, len_, 1, fp) != 1:             # <<<<<<<<<<<<<<
 *         raise_from_errno(IOError)
 *     return 0
 */
  }

  /* "s3ql/deltadump.pyx":124
 *     if fread_c(buf, len_, 1, fp) != 1:
 *         raise_from_errno(IOError)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef free(void * ptr):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":119
 *     return 0
 * 
 * cdef inline int fread(void * buf, size_t len_, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Call libc's fread() and raise exception on failure'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("s3ql.deltadump.fread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":126
 *     return 0
 * 
 * cdef free(void * ptr):             # <<<<<<<<<<<<<<
 *     '''Call libc.free()
 * 
 */

static PyObject *__pyx_f_4s3ql_9deltadump_free(void *__pyx_v_ptr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("free", 0);

  /* "s3ql/deltadump.pyx":133
 *     '''
 * 
 *     free_c(ptr)             # <<<<<<<<<<<<<<
 *     return None
 * 
 */
  free(__pyx_v_ptr);

  /* "s3ql/deltadump.pyx":134
 * 
 *     free_c(ptr)
 *     return None             # <<<<<<<<<<<<<<
 * 
 * cdef int raise_from_errno(err_class=OSError) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(Py_None);
  __pyx_r = Py_None;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":126
 *     return 0
 * 
 * cdef free(void * ptr):             # <<<<<<<<<<<<<<
 *     '''Call libc.free()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":136
 *     return None
 * 
 * cdef int raise_from_errno(err_class=OSError) except -1:             # <<<<<<<<<<<<<<
 *     '''Raise OSError for current errno value'''
 * 
 */

static int __pyx_f_4s3ql_9deltadump_raise_from_errno(struct __pyx_opt_args_4s3ql_9deltadump_raise_from_errno *__pyx_optional_args) {
  PyObject *__pyx_v_err_class = __pyx_k_;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("raise_from_errno", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_err_class = __pyx_optional_args->err_class;
    }
  }

  /* "s3ql/deltadump.pyx":139
 *     '''Raise OSError for current errno value'''
 * 
 *     raise err_class(errno, PyUnicode_FromString(strerror(errno)))             # <<<<<<<<<<<<<<
 * 
 * cdef int fclose(FILE * fp) except -1:
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(errno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyUnicode_FromString(strerror(errno)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_err_class);
  __pyx_t_4 = __pyx_v_err_class; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_5) {
    __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_t_3);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 139, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":136
 *     return None
 * 
 * cdef int raise_from_errno(err_class=OSError) except -1:             # <<<<<<<<<<<<<<
 *     '''Raise OSError for current errno value'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("s3ql.deltadump.raise_from_errno", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":141
 *     raise err_class(errno, PyUnicode_FromString(strerror(errno)))
 * 
 * cdef int fclose(FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Call libc.fclose() and raise exception on failure'''
 * 
 */

static int __pyx_f_4s3ql_9deltadump_fclose(FILE *__pyx_v_fp) {
  Py_ssize_t __pyx_v_off;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("fclose", 0);

  /* "s3ql/deltadump.pyx":149
 *     # important, so that we can safely reposition the fd position
 *     # below (which is necessary in case there is cached input data)
 *     if fflush(fp) != 0:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  __pyx_t_1 = ((fflush(__pyx_v_fp) != 0) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":150
 *     # below (which is necessary in case there is cached input data)
 *     if fflush(fp) != 0:
 *         raise_from_errno()             # <<<<<<<<<<<<<<
 * 
 *     # Reposition FD to position of FILE*, otherwise next read from FD will miss
 */
    __pyx_t_2 = __pyx_f_4s3ql_9deltadump_raise_from_errno(NULL); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 150, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":149
 *     # important, so that we can safely reposition the fd position
 *     # below (which is necessary in case there is cached input data)
 *     if fflush(fp) != 0:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":156
 *     # the same thing, but this does not seem to be documented so we don't rely
 *     # on it.
 *     off = ftell(fp)             # <<<<<<<<<<<<<<
 *     if off == -1:
 *         raise_from_errno()
 */
  __pyx_v_off = ftell(__pyx_v_fp);

  /* "s3ql/deltadump.pyx":157
 *     # on it.
 *     off = ftell(fp)
 *     if off == -1:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  __pyx_t_1 = ((__pyx_v_off == -1L) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":158
 *     off = ftell(fp)
 *     if off == -1:
 *         raise_from_errno()             # <<<<<<<<<<<<<<
 * 
 *     if lseek(fileno(fp), off, SEEK_SET) != off:
 */
    __pyx_t_2 = __pyx_f_4s3ql_9deltadump_raise_from_errno(NULL); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 158, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":157
 *     # on it.
 *     off = ftell(fp)
 *     if off == -1:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":160
 *         raise_from_errno()
 * 
 *     if lseek(fileno(fp), off, SEEK_SET) != off:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  __pyx_t_1 = ((lseek(fileno(__pyx_v_fp), __pyx_v_off, SEEK_SET) != __pyx_v_off) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":161
 * 
 *     if lseek(fileno(fp), off, SEEK_SET) != off:
 *         raise_from_errno()             # <<<<<<<<<<<<<<
 * 
 *     if fclose_c(fp) != 0:
 */
    __pyx_t_2 = __pyx_f_4s3ql_9deltadump_raise_from_errno(NULL); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 161, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":160
 *         raise_from_errno()
 * 
 *     if lseek(fileno(fp), off, SEEK_SET) != off:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":163
 *         raise_from_errno()
 * 
 *     if fclose_c(fp) != 0:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  __pyx_t_1 = ((fclose(__pyx_v_fp) != 0) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":164
 * 
 *     if fclose_c(fp) != 0:
 *         raise_from_errno()             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    __pyx_t_2 = __pyx_f_4s3ql_9deltadump_raise_from_errno(NULL); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 164, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":163
 *         raise_from_errno()
 * 
 *     if fclose_c(fp) != 0:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":166
 *         raise_from_errno()
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * cdef void * calloc(size_t cnt, size_t size) except NULL:
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":141
 *     raise err_class(errno, PyUnicode_FromString(strerror(errno)))
 * 
 * cdef int fclose(FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Call libc.fclose() and raise exception on failure'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("s3ql.deltadump.fclose", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":168
 *     return 0
 * 
 * cdef void * calloc(size_t cnt, size_t size) except NULL:             # <<<<<<<<<<<<<<
 *     '''Call libc.calloc and raise exception on failure'''
 * 
 */

static void *__pyx_f_4s3ql_9deltadump_calloc(size_t __pyx_v_cnt, size_t __pyx_v_size) {
  void *__pyx_v_ptr;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2;
  __Pyx_RefNannySetupContext("calloc", 0);

  /* "s3ql/deltadump.pyx":173
 *     cdef void * ptr
 * 
 *     ptr = calloc_c(cnt, size)             # <<<<<<<<<<<<<<
 * 
 *     if ptr is NULL:
 */
  __pyx_v_ptr = calloc(__pyx_v_cnt, __pyx_v_size);

  /* "s3ql/deltadump.pyx":175
 *     ptr = calloc_c(cnt, size)
 * 
 *     if ptr is NULL:             # <<<<<<<<<<<<<<
 *         PyErr_NoMemory()
 * 
 */
  __pyx_t_1 = ((__pyx_v_ptr == NULL) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":176
 * 
 *     if ptr is NULL:
 *         PyErr_NoMemory()             # <<<<<<<<<<<<<<
 * 
 *     return ptr
 */
    __pyx_t_2 = PyErr_NoMemory(); if (unlikely(__pyx_t_2 == NULL)) __PYX_ERR(0, 176, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":175
 *     ptr = calloc_c(cnt, size)
 * 
 *     if ptr is NULL:             # <<<<<<<<<<<<<<
 *         PyErr_NoMemory()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":178
 *         PyErr_NoMemory()
 * 
 *     return ptr             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_ptr;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":168
 *     return 0
 * 
 * cdef void * calloc(size_t cnt, size_t size) except NULL:             # <<<<<<<<<<<<<<
 *     '''Call libc.calloc and raise exception on failure'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("s3ql.deltadump.calloc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":181
 * 
 * 
 * cdef int SQLITE_CHECK_RC(int rc, int success, sqlite3* db) except -1:             # <<<<<<<<<<<<<<
 *     '''Raise correct exception if *rc* != *success*'''
 * 
 */

static int __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(int __pyx_v_rc, int __pyx_v_success, sqlite3 *__pyx_v_db) {
  PyObject *__pyx_v_exc = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("SQLITE_CHECK_RC", 0);

  /* "s3ql/deltadump.pyx":184
 *     '''Raise correct exception if *rc* != *success*'''
 * 
 *     if rc != success:             # <<<<<<<<<<<<<<
 *         exc = apsw.exceptionfor(rc)
 *         raise type(exc)(PyUnicode_FromString(sqlite3_errmsg(db)))
 */
  __pyx_t_1 = ((__pyx_v_rc != __pyx_v_success) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":185
 * 
 *     if rc != success:
 *         exc = apsw.exceptionfor(rc)             # <<<<<<<<<<<<<<
 *         raise type(exc)(PyUnicode_FromString(sqlite3_errmsg(db)))
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_apsw); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_exceptionfor); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_exc = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "s3ql/deltadump.pyx":186
 *     if rc != success:
 *         exc = apsw.exceptionfor(rc)
 *         raise type(exc)(PyUnicode_FromString(sqlite3_errmsg(db)))             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    __pyx_t_4 = PyUnicode_FromString(sqlite3_errmsg(__pyx_v_db)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_exc)));
    __pyx_t_6 = ((PyObject *)Py_TYPE(__pyx_v_exc)); __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 186, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":184
 *     '''Raise correct exception if *rc* != *success*'''
 * 
 *     if rc != success:             # <<<<<<<<<<<<<<
 *         exc = apsw.exceptionfor(rc)
 *         raise type(exc)(PyUnicode_FromString(sqlite3_errmsg(db)))
 */
  }

  /* "s3ql/deltadump.pyx":188
 *         raise type(exc)(PyUnicode_FromString(sqlite3_errmsg(db)))
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":181
 * 
 * 
 * cdef int SQLITE_CHECK_RC(int rc, int success, sqlite3* db) except -1:             # <<<<<<<<<<<<<<
 *     '''Raise correct exception if *rc* != *success*'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("s3ql.deltadump.SQLITE_CHECK_RC", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":191
 * 
 * 
 * cdef int prep_columns(columns, int** col_types_p, int** col_args_p) except -1:             # <<<<<<<<<<<<<<
 *     '''Allocate col_types and col_args, return number of columns
 * 
 */

static int __pyx_f_4s3ql_9deltadump_prep_columns(PyObject *__pyx_v_columns, int **__pyx_v_col_types_p, int **__pyx_v_col_args_p) {
  size_t __pyx_v_col_count;
  int *__pyx_v_col_types;
  int *__pyx_v_col_args;
  PyObject *__pyx_v_i = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  void *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  __Pyx_RefNannySetupContext("prep_columns", 0);

  /* "s3ql/deltadump.pyx":201
 *     cdef int *col_args
 * 
 *     col_count = <size_t> len(columns) # guaranteed positive             # <<<<<<<<<<<<<<
 *     col_types = < int *> calloc(col_count, sizeof(int))
 *     col_args = < int *> calloc(col_count, sizeof(int))
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_columns); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __pyx_v_col_count = ((size_t)__pyx_t_1);

  /* "s3ql/deltadump.pyx":202
 * 
 *     col_count = <size_t> len(columns) # guaranteed positive
 *     col_types = < int *> calloc(col_count, sizeof(int))             # <<<<<<<<<<<<<<
 *     col_args = < int *> calloc(col_count, sizeof(int))
 * 
 */
  __pyx_t_2 = __pyx_f_4s3ql_9deltadump_calloc(__pyx_v_col_count, (sizeof(int))); if (unlikely(__pyx_t_2 == NULL)) __PYX_ERR(0, 202, __pyx_L1_error)
  __pyx_v_col_types = ((int *)__pyx_t_2);

  /* "s3ql/deltadump.pyx":203
 *     col_count = <size_t> len(columns) # guaranteed positive
 *     col_types = < int *> calloc(col_count, sizeof(int))
 *     col_args = < int *> calloc(col_count, sizeof(int))             # <<<<<<<<<<<<<<
 * 
 *     # Initialize col_args and col_types
 */
  __pyx_t_2 = __pyx_f_4s3ql_9deltadump_calloc(__pyx_v_col_count, (sizeof(int))); if (unlikely(__pyx_t_2 == NULL)) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_v_col_args = ((int *)__pyx_t_2);

  /* "s3ql/deltadump.pyx":206
 * 
 *     # Initialize col_args and col_types
 *     for i in range(col_count):             # <<<<<<<<<<<<<<
 *         if columns[i][1] not in (BLOB, INTEGER, TIME):
 *             raise ValueError("Invalid type for column %d" % i)
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t(__pyx_v_col_count); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_1 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 206, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_1); __Pyx_INCREF(__pyx_t_3); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 206, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_5(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 206, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "s3ql/deltadump.pyx":207
 *     # Initialize col_args and col_types
 *     for i in range(col_count):
 *         if columns[i][1] not in (BLOB, INTEGER, TIME):             # <<<<<<<<<<<<<<
 *             raise ValueError("Invalid type for column %d" % i)
 *         col_types[i] = columns[i][1]
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_columns, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_3, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_BLOB); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_INTEGER); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_6, __pyx_t_8, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_9) {
    } else {
      __pyx_t_7 = __pyx_t_9;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_TIME); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_3, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_9 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __pyx_t_9;
    __pyx_L6_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = (__pyx_t_7 != 0);
    if (__pyx_t_9) {

      /* "s3ql/deltadump.pyx":208
 *     for i in range(col_count):
 *         if columns[i][1] not in (BLOB, INTEGER, TIME):
 *             raise ValueError("Invalid type for column %d" % i)             # <<<<<<<<<<<<<<
 *         col_types[i] = columns[i][1]
 * 
 */
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_Invalid_type_for_column_d, __pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 208, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 208, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_8, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 208, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 208, __pyx_L1_error)

      /* "s3ql/deltadump.pyx":207
 *     # Initialize col_args and col_types
 *     for i in range(col_count):
 *         if columns[i][1] not in (BLOB, INTEGER, TIME):             # <<<<<<<<<<<<<<
 *             raise ValueError("Invalid type for column %d" % i)
 *         col_types[i] = columns[i][1]
 */
    }

    /* "s3ql/deltadump.pyx":209
 *         if columns[i][1] not in (BLOB, INTEGER, TIME):
 *             raise ValueError("Invalid type for column %d" % i)
 *         col_types[i] = columns[i][1]             # <<<<<<<<<<<<<<
 * 
 *         if len(columns[i]) == 3:
 */
    __pyx_t_6 = PyObject_GetItem(__pyx_v_columns, __pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_t_6, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 209, __pyx_L1_error)
    (__pyx_v_col_types[__pyx_t_11]) = __pyx_t_10;

    /* "s3ql/deltadump.pyx":211
 *         col_types[i] = columns[i][1]
 * 
 *         if len(columns[i]) == 3:             # <<<<<<<<<<<<<<
 *             col_args[i] = columns[i][2]
 *         else:
 */
    __pyx_t_8 = PyObject_GetItem(__pyx_v_columns, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_11 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = ((__pyx_t_11 == 3) != 0);
    if (__pyx_t_9) {

      /* "s3ql/deltadump.pyx":212
 * 
 *         if len(columns[i]) == 3:
 *             col_args[i] = columns[i][2]             # <<<<<<<<<<<<<<
 *         else:
 *             col_args[i] = 0
 */
      __pyx_t_8 = PyObject_GetItem(__pyx_v_columns, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_8, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 212, __pyx_L1_error)
      (__pyx_v_col_args[__pyx_t_11]) = __pyx_t_10;

      /* "s3ql/deltadump.pyx":211
 *         col_types[i] = columns[i][1]
 * 
 *         if len(columns[i]) == 3:             # <<<<<<<<<<<<<<
 *             col_args[i] = columns[i][2]
 *         else:
 */
      goto __pyx_L9;
    }

    /* "s3ql/deltadump.pyx":214
 *             col_args[i] = columns[i][2]
 *         else:
 *             col_args[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     col_types_p[0] = col_types
 */
    /*else*/ {
      __pyx_t_11 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_11 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 214, __pyx_L1_error)
      (__pyx_v_col_args[__pyx_t_11]) = 0;
    }
    __pyx_L9:;

    /* "s3ql/deltadump.pyx":206
 * 
 *     # Initialize col_args and col_types
 *     for i in range(col_count):             # <<<<<<<<<<<<<<
 *         if columns[i][1] not in (BLOB, INTEGER, TIME):
 *             raise ValueError("Invalid type for column %d" % i)
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "s3ql/deltadump.pyx":216
 *             col_args[i] = 0
 * 
 *     col_types_p[0] = col_types             # <<<<<<<<<<<<<<
 *     col_args_p[0] = col_args
 * 
 */
  (__pyx_v_col_types_p[0]) = __pyx_v_col_types;

  /* "s3ql/deltadump.pyx":217
 * 
 *     col_types_p[0] = col_types
 *     col_args_p[0] = col_args             # <<<<<<<<<<<<<<
 * 
 *     # We can safely assume that this fits into an int
 */
  (__pyx_v_col_args_p[0]) = __pyx_v_col_args;

  /* "s3ql/deltadump.pyx":220
 * 
 *     # We can safely assume that this fits into an int
 *     return <int> col_count             # <<<<<<<<<<<<<<
 * 
 * cdef FILE* dup_to_fp(fh, const_char* mode) except NULL:
 */
  __pyx_r = ((int)__pyx_v_col_count);
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":191
 * 
 * 
 * cdef int prep_columns(columns, int** col_types_p, int** col_args_p) except -1:             # <<<<<<<<<<<<<<
 *     '''Allocate col_types and col_args, return number of columns
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("s3ql.deltadump.prep_columns", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":222
 *     return <int> col_count
 * 
 * cdef FILE* dup_to_fp(fh, const_char* mode) except NULL:             # <<<<<<<<<<<<<<
 *     '''Duplicate fd from *fh* and open as FILE*'''
 * 
 */

static FILE *__pyx_f_4s3ql_9deltadump_dup_to_fp(PyObject *__pyx_v_fh, const char *__pyx_v_mode) {
  int __pyx_v_fd;
  FILE *__pyx_v_fp;
  FILE *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("dup_to_fp", 0);

  /* "s3ql/deltadump.pyx":227
 *     cdef int fd
 * 
 *     fd = dup(fh.fileno())             # <<<<<<<<<<<<<<
 *     if fd == -1:
 *         raise_from_errno()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_fh, __pyx_n_s_fileno); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_fd = dup(__pyx_t_4);

  /* "s3ql/deltadump.pyx":228
 * 
 *     fd = dup(fh.fileno())
 *     if fd == -1:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  __pyx_t_5 = ((__pyx_v_fd == -1L) != 0);
  if (__pyx_t_5) {

    /* "s3ql/deltadump.pyx":229
 *     fd = dup(fh.fileno())
 *     if fd == -1:
 *         raise_from_errno()             # <<<<<<<<<<<<<<
 * 
 *     fp = fdopen(fd, mode)
 */
    __pyx_t_4 = __pyx_f_4s3ql_9deltadump_raise_from_errno(NULL); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 229, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":228
 * 
 *     fd = dup(fh.fileno())
 *     if fd == -1:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":231
 *         raise_from_errno()
 * 
 *     fp = fdopen(fd, mode)             # <<<<<<<<<<<<<<
 *     if fp == NULL:
 *         raise_from_errno()
 */
  __pyx_v_fp = fdopen(__pyx_v_fd, __pyx_v_mode);

  /* "s3ql/deltadump.pyx":232
 * 
 *     fp = fdopen(fd, mode)
 *     if fp == NULL:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  __pyx_t_5 = ((__pyx_v_fp == NULL) != 0);
  if (__pyx_t_5) {

    /* "s3ql/deltadump.pyx":233
 *     fp = fdopen(fd, mode)
 *     if fp == NULL:
 *         raise_from_errno()             # <<<<<<<<<<<<<<
 * 
 *     return fp
 */
    __pyx_t_4 = __pyx_f_4s3ql_9deltadump_raise_from_errno(NULL); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 233, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":232
 * 
 *     fp = fdopen(fd, mode)
 *     if fp == NULL:             # <<<<<<<<<<<<<<
 *         raise_from_errno()
 * 
 */
  }

  /* "s3ql/deltadump.pyx":235
 *         raise_from_errno()
 * 
 *     return fp             # <<<<<<<<<<<<<<
 * 
 * def check_sqlite():
 */
  __pyx_r = __pyx_v_fp;
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":222
 *     return <int> col_count
 * 
 * cdef FILE* dup_to_fp(fh, const_char* mode) except NULL:             # <<<<<<<<<<<<<<
 *     '''Duplicate fd from *fh* and open as FILE*'''
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("s3ql.deltadump.dup_to_fp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":237
 *     return fp
 * 
 * def check_sqlite():             # <<<<<<<<<<<<<<
 *     '''Check if deltadump and apsw module use compatible SQLite code.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_1check_sqlite(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4s3ql_9deltadump_check_sqlite[] = "check_sqlite()\nCheck if deltadump and apsw module use compatible SQLite code.\n\n    This functions look at versions and compile options of the SQLite\n    code used by the *apsw* module and the *deltadump* module. If they\n    do not match exactly, a `RuntimeError` is raised.\n\n    Only if both modules use the same SQLite version compiled with the\n    same options can the database object be shared between *apsw* and\n    *deltadump*.\n    ";
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_1check_sqlite = {"check_sqlite", (PyCFunction)__pyx_pw_4s3ql_9deltadump_1check_sqlite, METH_NOARGS, __pyx_doc_4s3ql_9deltadump_check_sqlite};
static PyObject *__pyx_pw_4s3ql_9deltadump_1check_sqlite(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_sqlite (wrapper)", 0);
  __pyx_r = __pyx_pf_4s3ql_9deltadump_check_sqlite(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4s3ql_9deltadump_check_sqlite(CYTHON_UNUSED PyObject *__pyx_self) {
  const char *__pyx_v_buf;
  PyObject *__pyx_v_apsw_sqlite_version = NULL;
  PyObject *__pyx_v_s3ql_sqlite_version = NULL;
  PyObject *__pyx_v_apsw_sqlite_options = NULL;
  PyObject *__pyx_v_s3ql_sqlite_options = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("check_sqlite", 0);

  /* "s3ql/deltadump.pyx":251
 *     cdef const_char *buf
 * 
 *     apsw_sqlite_version = apsw.sqlitelibversion()             # <<<<<<<<<<<<<<
 *     s3ql_sqlite_version = PyUnicode_FromString(sqlite3_libversion())
 *     log.debug('apsw sqlite version: %s, '
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_apsw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sqlitelibversion); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_apsw_sqlite_version = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":252
 * 
 *     apsw_sqlite_version = apsw.sqlitelibversion()
 *     s3ql_sqlite_version = PyUnicode_FromString(sqlite3_libversion())             # <<<<<<<<<<<<<<
 *     log.debug('apsw sqlite version: %s, '
 *               's3ql sqlite version: %s',
 */
  __pyx_t_1 = PyUnicode_FromString(sqlite3_libversion()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s3ql_sqlite_version = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":253
 *     apsw_sqlite_version = apsw.sqlitelibversion()
 *     s3ql_sqlite_version = PyUnicode_FromString(sqlite3_libversion())
 *     log.debug('apsw sqlite version: %s, '             # <<<<<<<<<<<<<<
 *               's3ql sqlite version: %s',
 *               apsw_sqlite_version,
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "s3ql/deltadump.pyx":256
 *               's3ql sqlite version: %s',
 *               apsw_sqlite_version,
 *               s3ql_sqlite_version)             # <<<<<<<<<<<<<<
 *     if apsw_sqlite_version != s3ql_sqlite_version:
 *         raise RuntimeError('SQLite version mismatch between APSW and S3QL '
 */
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_5 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_3) {
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3); __pyx_t_3 = NULL;
  }
  __Pyx_INCREF(__pyx_kp_u_apsw_sqlite_version_s_s3ql_sqlit);
  __Pyx_GIVEREF(__pyx_kp_u_apsw_sqlite_version_s_s3ql_sqlit);
  PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_4, __pyx_kp_u_apsw_sqlite_version_s_s3ql_sqlit);
  __Pyx_INCREF(__pyx_v_apsw_sqlite_version);
  __Pyx_GIVEREF(__pyx_v_apsw_sqlite_version);
  PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_4, __pyx_v_apsw_sqlite_version);
  __Pyx_INCREF(__pyx_v_s3ql_sqlite_version);
  __Pyx_GIVEREF(__pyx_v_s3ql_sqlite_version);
  PyTuple_SET_ITEM(__pyx_t_5, 2+__pyx_t_4, __pyx_v_s3ql_sqlite_version);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":257
 *               apsw_sqlite_version,
 *               s3ql_sqlite_version)
 *     if apsw_sqlite_version != s3ql_sqlite_version:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('SQLite version mismatch between APSW and S3QL '
 *                            '(%s vs %s)' % (apsw_sqlite_version, s3ql_sqlite_version))
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_apsw_sqlite_version, __pyx_v_s3ql_sqlite_version, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "s3ql/deltadump.pyx":259
 *     if apsw_sqlite_version != s3ql_sqlite_version:
 *         raise RuntimeError('SQLite version mismatch between APSW and S3QL '
 *                            '(%s vs %s)' % (apsw_sqlite_version, s3ql_sqlite_version))             # <<<<<<<<<<<<<<
 * 
 *     apsw_sqlite_options = set(apsw.compile_options)
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_apsw_sqlite_version);
    __Pyx_GIVEREF(__pyx_v_apsw_sqlite_version);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_apsw_sqlite_version);
    __Pyx_INCREF(__pyx_v_s3ql_sqlite_version);
    __Pyx_GIVEREF(__pyx_v_s3ql_sqlite_version);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_s3ql_sqlite_version);
    __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_SQLite_version_mismatch_between, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "s3ql/deltadump.pyx":258
 *               s3ql_sqlite_version)
 *     if apsw_sqlite_version != s3ql_sqlite_version:
 *         raise RuntimeError('SQLite version mismatch between APSW and S3QL '             # <<<<<<<<<<<<<<
 *                            '(%s vs %s)' % (apsw_sqlite_version, s3ql_sqlite_version))
 * 
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 258, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":257
 *               apsw_sqlite_version,
 *               s3ql_sqlite_version)
 *     if apsw_sqlite_version != s3ql_sqlite_version:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('SQLite version mismatch between APSW and S3QL '
 *                            '(%s vs %s)' % (apsw_sqlite_version, s3ql_sqlite_version))
 */
  }

  /* "s3ql/deltadump.pyx":261
 *                            '(%s vs %s)' % (apsw_sqlite_version, s3ql_sqlite_version))
 * 
 *     apsw_sqlite_options = set(apsw.compile_options)             # <<<<<<<<<<<<<<
 *     s3ql_sqlite_options = set()
 *     for idx in itertools.count(0):
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_apsw); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_compile_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_apsw_sqlite_options = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "s3ql/deltadump.pyx":262
 * 
 *     apsw_sqlite_options = set(apsw.compile_options)
 *     s3ql_sqlite_options = set()             # <<<<<<<<<<<<<<
 *     for idx in itertools.count(0):
 *         buf = sqlite3_compileoption_get(idx)
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_s3ql_sqlite_options = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "s3ql/deltadump.pyx":263
 *     apsw_sqlite_options = set(apsw.compile_options)
 *     s3ql_sqlite_options = set()
 *     for idx in itertools.count(0):             # <<<<<<<<<<<<<<
 *         buf = sqlite3_compileoption_get(idx)
 *         if buf is NULL:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_itertools); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_count); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1); __pyx_t_4 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 263, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 263, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 263, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 263, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 263, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "s3ql/deltadump.pyx":264
 *     s3ql_sqlite_options = set()
 *     for idx in itertools.count(0):
 *         buf = sqlite3_compileoption_get(idx)             # <<<<<<<<<<<<<<
 *         if buf is NULL:
 *             break
 */
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_idx); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 264, __pyx_L1_error)
    __pyx_v_buf = sqlite3_compileoption_get(__pyx_t_8);

    /* "s3ql/deltadump.pyx":265
 *     for idx in itertools.count(0):
 *         buf = sqlite3_compileoption_get(idx)
 *         if buf is NULL:             # <<<<<<<<<<<<<<
 *             break
 *         s3ql_sqlite_options.add(PyUnicode_FromString(buf))
 */
    __pyx_t_6 = ((__pyx_v_buf == NULL) != 0);
    if (__pyx_t_6) {

      /* "s3ql/deltadump.pyx":266
 *         buf = sqlite3_compileoption_get(idx)
 *         if buf is NULL:
 *             break             # <<<<<<<<<<<<<<
 *         s3ql_sqlite_options.add(PyUnicode_FromString(buf))
 * 
 */
      goto __pyx_L5_break;

      /* "s3ql/deltadump.pyx":265
 *     for idx in itertools.count(0):
 *         buf = sqlite3_compileoption_get(idx)
 *         if buf is NULL:             # <<<<<<<<<<<<<<
 *             break
 *         s3ql_sqlite_options.add(PyUnicode_FromString(buf))
 */
    }

    /* "s3ql/deltadump.pyx":267
 *         if buf is NULL:
 *             break
 *         s3ql_sqlite_options.add(PyUnicode_FromString(buf))             # <<<<<<<<<<<<<<
 * 
 *     log.debug('apsw sqlite compile options: %s, '
 */
    __pyx_t_2 = PyUnicode_FromString(__pyx_v_buf); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 267, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_9 = PySet_Add(__pyx_v_s3ql_sqlite_options, __pyx_t_2); if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 267, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "s3ql/deltadump.pyx":263
 *     apsw_sqlite_options = set(apsw.compile_options)
 *     s3ql_sqlite_options = set()
 *     for idx in itertools.count(0):             # <<<<<<<<<<<<<<
 *         buf = sqlite3_compileoption_get(idx)
 *         if buf is NULL:
 */
  }
  __pyx_L5_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":269
 *         s3ql_sqlite_options.add(PyUnicode_FromString(buf))
 * 
 *     log.debug('apsw sqlite compile options: %s, '             # <<<<<<<<<<<<<<
 *               's3ql sqlite compile options: %s',
 *               apsw_sqlite_options,
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "s3ql/deltadump.pyx":272
 *               's3ql sqlite compile options: %s',
 *               apsw_sqlite_options,
 *               s3ql_sqlite_options)             # <<<<<<<<<<<<<<
 *     if apsw_sqlite_options != s3ql_sqlite_options:
 *         raise RuntimeError('SQLite code used by APSW was compiled with different '
 */
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_4 = 1;
    }
  }
  __pyx_t_3 = PyTuple_New(3+__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_2) {
    __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2); __pyx_t_2 = NULL;
  }
  __Pyx_INCREF(__pyx_kp_u_apsw_sqlite_compile_options_s_s3);
  __Pyx_GIVEREF(__pyx_kp_u_apsw_sqlite_compile_options_s_s3);
  PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_4, __pyx_kp_u_apsw_sqlite_compile_options_s_s3);
  __Pyx_INCREF(__pyx_v_apsw_sqlite_options);
  __Pyx_GIVEREF(__pyx_v_apsw_sqlite_options);
  PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_4, __pyx_v_apsw_sqlite_options);
  __Pyx_INCREF(__pyx_v_s3ql_sqlite_options);
  __Pyx_GIVEREF(__pyx_v_s3ql_sqlite_options);
  PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_4, __pyx_v_s3ql_sqlite_options);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":273
 *               apsw_sqlite_options,
 *               s3ql_sqlite_options)
 *     if apsw_sqlite_options != s3ql_sqlite_options:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('SQLite code used by APSW was compiled with different '
 *                            'options than SQLite code available to S3QL! '
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_apsw_sqlite_options, __pyx_v_s3ql_sqlite_options, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "s3ql/deltadump.pyx":277
 *                            'options than SQLite code available to S3QL! '
 *                            'Differing settings: + %s, - %s' %
 *                            (apsw_sqlite_options - s3ql_sqlite_options,             # <<<<<<<<<<<<<<
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 */
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_apsw_sqlite_options, __pyx_v_s3ql_sqlite_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "s3ql/deltadump.pyx":278
 *                            'Differing settings: + %s, - %s' %
 *                            (apsw_sqlite_options - s3ql_sqlite_options,
 *                            s3ql_sqlite_options - apsw_sqlite_options))             # <<<<<<<<<<<<<<
 * 
 * def dump_table(table, order, columns, db, fh):
 */
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_s3ql_sqlite_options, __pyx_v_apsw_sqlite_options); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "s3ql/deltadump.pyx":277
 *                            'options than SQLite code available to S3QL! '
 *                            'Differing settings: + %s, - %s' %
 *                            (apsw_sqlite_options - s3ql_sqlite_options,             # <<<<<<<<<<<<<<
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 */
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __pyx_t_1 = 0;
    __pyx_t_5 = 0;

    /* "s3ql/deltadump.pyx":276
 *         raise RuntimeError('SQLite code used by APSW was compiled with different '
 *                            'options than SQLite code available to S3QL! '
 *                            'Differing settings: + %s, - %s' %             # <<<<<<<<<<<<<<
 *                            (apsw_sqlite_options - s3ql_sqlite_options,
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 */
    __pyx_t_5 = PyUnicode_Format(__pyx_kp_u_SQLite_code_used_by_APSW_was_com, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 276, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "s3ql/deltadump.pyx":274
 *               s3ql_sqlite_options)
 *     if apsw_sqlite_options != s3ql_sqlite_options:
 *         raise RuntimeError('SQLite code used by APSW was compiled with different '             # <<<<<<<<<<<<<<
 *                            'options than SQLite code available to S3QL! '
 *                            'Differing settings: + %s, - %s' %
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 274, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":273
 *               apsw_sqlite_options,
 *               s3ql_sqlite_options)
 *     if apsw_sqlite_options != s3ql_sqlite_options:             # <<<<<<<<<<<<<<
 *         raise RuntimeError('SQLite code used by APSW was compiled with different '
 *                            'options than SQLite code available to S3QL! '
 */
  }

  /* "s3ql/deltadump.pyx":237
 *     return fp
 * 
 * def check_sqlite():             # <<<<<<<<<<<<<<
 *     '''Check if deltadump and apsw module use compatible SQLite code.
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("s3ql.deltadump.check_sqlite", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_apsw_sqlite_version);
  __Pyx_XDECREF(__pyx_v_s3ql_sqlite_version);
  __Pyx_XDECREF(__pyx_v_apsw_sqlite_options);
  __Pyx_XDECREF(__pyx_v_s3ql_sqlite_options);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":280
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 * def dump_table(table, order, columns, db, fh):             # <<<<<<<<<<<<<<
 *     '''Dump *columns* of *table* into *fh*
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_3dump_table(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4s3ql_9deltadump_2dump_table[] = "dump_table(table, order, columns, db, fh)\nDump *columns* of *table* into *fh*\n\n    *order* specifies the order in which the rows are written and must be a\n    string that can be inserted after the \"ORDER BY\" clause in an SQL SELECT\n    statement.\n\n    *db* is an `s3ql.Connection` instance for the database.\n\n    *columns* must a list of 3-tuples, one for each column that should be\n    stored. The first element of the tuple must contain the column name and the\n    second element the type of data stored in the column (`INTEGER`, `TIME`\n    or `BLOB`). Times will be converted to nanosecond integers.\n\n    For integers and seconds, the third tuple element specifies the expected\n    change of the values between rows. For blobs it can be either zero\n    (indicating variable length columns) or an integer specifying the length of\n    the column values in bytes.\n\n    This function will open a separate connection to the database, so\n    the *db* connection should not be in EXCLUSIVE locking mode.\n    (Using a separate connection avoids the requirement on the *apsw*\n    and *deltadump* modules be linked against against binary\n    compatible SQLite libraries).\n    ";
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_3dump_table = {"dump_table", (PyCFunction)__pyx_pw_4s3ql_9deltadump_3dump_table, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4s3ql_9deltadump_2dump_table};
static PyObject *__pyx_pw_4s3ql_9deltadump_3dump_table(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_table = 0;
  PyObject *__pyx_v_order = 0;
  PyObject *__pyx_v_columns = 0;
  PyObject *__pyx_v_db = 0;
  PyObject *__pyx_v_fh = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dump_table (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_table,&__pyx_n_s_order,&__pyx_n_s_columns,&__pyx_n_s_db,&__pyx_n_s_fh,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_table)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_order)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dump_table", 1, 5, 5, 1); __PYX_ERR(0, 280, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_columns)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dump_table", 1, 5, 5, 2); __PYX_ERR(0, 280, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_db)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dump_table", 1, 5, 5, 3); __PYX_ERR(0, 280, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_fh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dump_table", 1, 5, 5, 4); __PYX_ERR(0, 280, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dump_table") < 0)) __PYX_ERR(0, 280, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_table = values[0];
    __pyx_v_order = values[1];
    __pyx_v_columns = values[2];
    __pyx_v_db = values[3];
    __pyx_v_fh = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dump_table", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 280, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("s3ql.deltadump.dump_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4s3ql_9deltadump_2dump_table(__pyx_self, __pyx_v_table, __pyx_v_order, __pyx_v_columns, __pyx_v_db, __pyx_v_fh);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":328
 *                                         SQLITE_OPEN_READONLY, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10dump_table_lambda = {"lambda", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10dump_table_lambda, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":329
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),
 *                                             SQLITE_OK, sqlite3_db))             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 *                         SQLITE_OK, sqlite3_db)
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_close(__pyx_cur_scope->__pyx_v_sqlite3_db), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 328, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":328
 *                                         SQLITE_OPEN_READONLY, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":335
 *         # Get FILE* for buffered reading from *fh*
 *         fp = dup_to_fp(fh, b'wb')
 *         cm.callback(lambda: fclose(fp))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate col_args and col_types
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_1lambda1(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10dump_table_1lambda1 = {"lambda1", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10dump_table_1lambda1, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_1lambda1(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda1 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda1(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda1(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda1", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_fclose(__pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table.lambda1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":339
 *         # Allocate col_args and col_types
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(col_types))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_2lambda2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10dump_table_2lambda2 = {"lambda2", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10dump_table_2lambda2, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_2lambda2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda2 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda2(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda2(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda2", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_col_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":340
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))
 *         cm.callback(lambda: free(col_types))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate int64_prev
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_3lambda3(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10dump_table_3lambda3 = {"lambda3", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10dump_table_3lambda3, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_3lambda3(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda3 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda3(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda3(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda3", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_col_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table.lambda3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":344
 *         # Allocate int64_prev
 *         int64_prev = <int64_t*> calloc(<size_t> len(columns), sizeof(int64_t))
 *         cm.callback(lambda: free(int64_prev))             # <<<<<<<<<<<<<<
 * 
 *         # Prepare statement
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_4lambda4(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10dump_table_4lambda4 = {"lambda4", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10dump_table_4lambda4, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_4lambda4(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda4 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda4(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda4(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda4", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_int64_prev); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table.lambda4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":352
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_5lambda5(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10dump_table_5lambda5 = {"lambda5", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10dump_table_5lambda5, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10dump_table_5lambda5(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda5 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda5(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda5(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda5", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":353
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),
 *                                             SQLITE_OK, sqlite3_db))             # <<<<<<<<<<<<<<
 * 
 *         row_count = db.get_val("SELECT COUNT(rowid) FROM %s" % table)
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_finalize(__pyx_cur_scope->__pyx_v_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 352, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":352
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table.lambda5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":280
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 * def dump_table(table, order, columns, db, fh):             # <<<<<<<<<<<<<<
 *     '''Dump *columns* of *table* into *fh*
 * 
 */

static PyObject *__pyx_pf_4s3ql_9deltadump_2dump_table(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_table, PyObject *__pyx_v_order, PyObject *__pyx_v_columns, PyObject *__pyx_v_db, PyObject *__pyx_v_fh) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_cur_scope;
  int __pyx_v_col_count;
  int __pyx_v_rc;
  int __pyx_v_i;
  size_t __pyx_v_len_;
  int64_t __pyx_v_int64;
  int64_t __pyx_v_tmp;
  const void *__pyx_v_buf;
  int64_t __pyx_v_row_count;
  PyObject *__pyx_v_cm = NULL;
  PyObject *__pyx_v_dbfile_b = NULL;
  PyObject *__pyx_v_col_names = NULL;
  PyObject *__pyx_v_query = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  const char *__pyx_t_12;
  int __pyx_t_13;
  FILE *__pyx_t_14;
  void *__pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  char *__pyx_t_17;
  int64_t __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_t_22;
  __Pyx_RefNannySetupContext("dump_table", 0);
  __pyx_cur_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *)__pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct__dump_table(__pyx_ptype_4s3ql_9deltadump___pyx_scope_struct__dump_table, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);

  /* "s3ql/deltadump.pyx":318
 *     cdef int64_t row_count
 * 
 *     if db.file == ':memory:':             # <<<<<<<<<<<<<<
 *         raise ValueError("Can't access in-memory databases")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u_memory, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":319
 * 
 *     if db.file == ':memory:':
 *         raise ValueError("Can't access in-memory databases")             # <<<<<<<<<<<<<<
 * 
 *     with ExitStack() as cm:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 319, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":318
 *     cdef int64_t row_count
 * 
 *     if db.file == ':memory:':             # <<<<<<<<<<<<<<
 *         raise ValueError("Can't access in-memory databases")
 * 
 */
  }

  /* "s3ql/deltadump.pyx":321
 *         raise ValueError("Can't access in-memory databases")
 * 
 *     with ExitStack() as cm:             # <<<<<<<<<<<<<<
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 */
  /*with:*/ {
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExitStack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 321, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L4_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {
          __pyx_v_cm = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":323
 *     with ExitStack() as cm:
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)             # <<<<<<<<<<<<<<
 *         dbfile_b = db.file.encode(sys.getfilesystemencoding(), 'surrogateescape')
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_10 = 1;
            }
          }
          __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 323, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_Opening_connection_to_s);
          __Pyx_GIVEREF(__pyx_kp_u_Opening_connection_to_s);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_kp_u_Opening_connection_to_s);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 323, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":324
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 *         dbfile_b = db.file.encode(sys.getfilesystemencoding(), 'surrogateescape')             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,
 *                                         SQLITE_OPEN_READONLY, NULL),
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 324, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getfilesystemencoding); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 324, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 324, __pyx_L8_error)
          }
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_10 = 1;
            }
          }
          __pyx_t_1 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_10, __pyx_t_3);
          __Pyx_INCREF(__pyx_n_u_surrogateescape);
          __Pyx_GIVEREF(__pyx_n_u_surrogateescape);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_10, __pyx_n_u_surrogateescape);
          __pyx_t_3 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 324, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_dbfile_b = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":325
 *         log.debug('Opening connection to %s', db.file)
 *         dbfile_b = db.file.encode(sys.getfilesystemencoding(), 'surrogateescape')
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,             # <<<<<<<<<<<<<<
 *                                         SQLITE_OPEN_READONLY, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __pyx_t_12 = __Pyx_PyObject_AsString(__pyx_v_dbfile_b); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 325, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":327
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,
 *                                         SQLITE_OPEN_READONLY, NULL),
 *                         SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),
 *                                             SQLITE_OK, sqlite3_db))
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_open_v2(__pyx_t_12, (&__pyx_cur_scope->__pyx_v_sqlite3_db), SQLITE_OPEN_READONLY, NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 325, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":328
 *                                         SQLITE_OPEN_READONLY, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 328, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10dump_table_lambda, 0, __pyx_n_s_dump_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 328, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 328, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":330
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),             # <<<<<<<<<<<<<<
 *                         SQLITE_OK, sqlite3_db)
 * 
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_extended_result_codes(__pyx_cur_scope->__pyx_v_sqlite3_db, 1), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 330, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":334
 * 
 *         # Get FILE* for buffered reading from *fh*
 *         fp = dup_to_fp(fh, b'wb')             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: fclose(fp))
 * 
 */
          __pyx_t_14 = __pyx_f_4s3ql_9deltadump_dup_to_fp(__pyx_v_fh, ((const char *)"wb")); if (unlikely(__pyx_t_14 == NULL)) __PYX_ERR(0, 334, __pyx_L8_error)
          __pyx_cur_scope->__pyx_v_fp = __pyx_t_14;

          /* "s3ql/deltadump.pyx":335
 *         # Get FILE* for buffered reading from *fh*
 *         fp = dup_to_fp(fh, b'wb')
 *         cm.callback(lambda: fclose(fp))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate col_args and col_types
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 335, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10dump_table_1lambda1, 0, __pyx_n_s_dump_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_1 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 335, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 335, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":338
 * 
 *         # Allocate col_args and col_types
 *         col_count = prep_columns(columns, &col_types, &col_args)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(col_args))
 *         cm.callback(lambda: free(col_types))
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_prep_columns(__pyx_v_columns, (&__pyx_cur_scope->__pyx_v_col_types), (&__pyx_cur_scope->__pyx_v_col_args)); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 338, __pyx_L8_error)
          __pyx_v_col_count = __pyx_t_13;

          /* "s3ql/deltadump.pyx":339
 *         # Allocate col_args and col_types
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(col_types))
 * 
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 339, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10dump_table_2lambda2, 0, __pyx_n_s_dump_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 339, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_6) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 339, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 339, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":340
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))
 *         cm.callback(lambda: free(col_types))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate int64_prev
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 340, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10dump_table_3lambda3, 0, __pyx_n_s_dump_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 340, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 340, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 340, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":343
 * 
 *         # Allocate int64_prev
 *         int64_prev = <int64_t*> calloc(<size_t> len(columns), sizeof(int64_t))             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(int64_prev))
 * 
 */
          __pyx_t_10 = PyObject_Length(__pyx_v_columns); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 343, __pyx_L8_error)
          __pyx_t_15 = __pyx_f_4s3ql_9deltadump_calloc(((size_t)__pyx_t_10), (sizeof(int64_t))); if (unlikely(__pyx_t_15 == NULL)) __PYX_ERR(0, 343, __pyx_L8_error)
          __pyx_cur_scope->__pyx_v_int64_prev = ((int64_t *)__pyx_t_15);

          /* "s3ql/deltadump.pyx":344
 *         # Allocate int64_prev
 *         int64_prev = <int64_t*> calloc(<size_t> len(columns), sizeof(int64_t))
 *         cm.callback(lambda: free(int64_prev))             # <<<<<<<<<<<<<<
 * 
 *         # Prepare statement
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 344, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10dump_table_4lambda4, 0, __pyx_n_s_dump_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 344, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_1 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 344, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 344, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 344, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":347
 * 
 *         # Prepare statement
 *         col_names = [ x[0] for x in columns ]             # <<<<<<<<<<<<<<
 *         query = ("SELECT %s FROM %s ORDER BY %s " %
 *                  (', '.join(col_names), table, order)).encode('utf-8')
 */
          { /* enter inner scope */
            PyObject *__pyx_7genexpr__pyx_v_x = NULL;
            __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 347, __pyx_L18_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (likely(PyList_CheckExact(__pyx_v_columns)) || PyTuple_CheckExact(__pyx_v_columns)) {
              __pyx_t_11 = __pyx_v_columns; __Pyx_INCREF(__pyx_t_11); __pyx_t_10 = 0;
              __pyx_t_16 = NULL;
            } else {
              __pyx_t_10 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_v_columns); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 347, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_16 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 347, __pyx_L18_error)
            }
            for (;;) {
              if (likely(!__pyx_t_16)) {
                if (likely(PyList_CheckExact(__pyx_t_11))) {
                  if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_11)) break;
                  #if CYTHON_COMPILING_IN_CPYTHON
                  __pyx_t_3 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 347, __pyx_L18_error)
                  #else
                  __pyx_t_3 = PySequence_ITEM(__pyx_t_11, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L18_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  #endif
                } else {
                  if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
                  #if CYTHON_COMPILING_IN_CPYTHON
                  __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_10); __Pyx_INCREF(__pyx_t_3); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 347, __pyx_L18_error)
                  #else
                  __pyx_t_3 = PySequence_ITEM(__pyx_t_11, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L18_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  #endif
                }
              } else {
                __pyx_t_3 = __pyx_t_16(__pyx_t_11);
                if (unlikely(!__pyx_t_3)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 347, __pyx_L18_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_3);
              }
              __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_x, __pyx_t_3);
              __pyx_t_3 = 0;
              __pyx_t_3 = __Pyx_GetItemInt(__pyx_7genexpr__pyx_v_x, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 347, __pyx_L18_error)
              __Pyx_GOTREF(__pyx_t_3);
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 347, __pyx_L18_error)
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_7genexpr__pyx_v_x);
            goto __pyx_L21_exit_scope;
            __pyx_L18_error:;
            __Pyx_XDECREF(__pyx_7genexpr__pyx_v_x);
            goto __pyx_L8_error;
            __pyx_L21_exit_scope:;
          } /* exit inner scope */
          __pyx_v_col_names = ((PyObject*)__pyx_t_4);
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":349
 *         col_names = [ x[0] for x in columns ]
 *         query = ("SELECT %s FROM %s ORDER BY %s " %
 *                  (', '.join(col_names), table, order)).encode('utf-8')             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __pyx_t_4 = PyUnicode_Join(__pyx_kp_u__4, __pyx_v_col_names); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 349, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 349, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_4);
          __Pyx_INCREF(__pyx_v_table);
          __Pyx_GIVEREF(__pyx_v_table);
          PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_table);
          __Pyx_INCREF(__pyx_v_order);
          __Pyx_GIVEREF(__pyx_v_order);
          PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_v_order);
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":348
 *         # Prepare statement
 *         col_names = [ x[0] for x in columns ]
 *         query = ("SELECT %s FROM %s ORDER BY %s " %             # <<<<<<<<<<<<<<
 *                  (', '.join(col_names), table, order)).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 */
          __pyx_t_4 = PyUnicode_Format(__pyx_kp_u_SELECT_s_FROM_s_ORDER_BY_s, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 348, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "s3ql/deltadump.pyx":349
 *         col_names = [ x[0] for x in columns ]
 *         query = ("SELECT %s FROM %s ORDER BY %s " %
 *                  (', '.join(col_names), table, order)).encode('utf-8')             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __pyx_t_11 = PyUnicode_AsUTF8String(((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 349, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_v_query = __pyx_t_11;
          __pyx_t_11 = 0;

          /* "s3ql/deltadump.pyx":350
 *         query = ("SELECT %s FROM %s ORDER BY %s " %
 *                  (', '.join(col_names), table, order)).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),             # <<<<<<<<<<<<<<
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),
 */
          __pyx_t_17 = __Pyx_PyObject_AsString(__pyx_v_query); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 350, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":351
 *                  (', '.join(col_names), table, order)).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),
 *                                             SQLITE_OK, sqlite3_db))
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_prepare_v2(__pyx_cur_scope->__pyx_v_sqlite3_db, __pyx_t_17, -1, (&__pyx_cur_scope->__pyx_v_stmt), NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 350, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":352
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 * 
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10dump_table_5lambda5, 0, __pyx_n_s_dump_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_6 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_6) {
            __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 352, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __Pyx_GOTREF(__pyx_t_11);
          } else {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_GIVEREF(__pyx_t_3);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_3);
            __pyx_t_3 = 0;
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_1, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 352, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "s3ql/deltadump.pyx":355
 *                                             SQLITE_OK, sqlite3_db))
 * 
 *         row_count = db.get_val("SELECT COUNT(rowid) FROM %s" % table)             # <<<<<<<<<<<<<<
 *         log.debug('dump_table(%s): writing %d rows', table, row_count)
 *         write_integer(row_count, fp)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_get_val); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 355, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_SELECT_COUNT_rowid_FROM_s, __pyx_v_table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 355, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_11);
          } else {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 355, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 355, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_18 = __Pyx_PyInt_As_int64_t(__pyx_t_11); if (unlikely((__pyx_t_18 == (int64_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 355, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_row_count = __pyx_t_18;

          /* "s3ql/deltadump.pyx":356
 * 
 *         row_count = db.get_val("SELECT COUNT(rowid) FROM %s" % table)
 *         log.debug('dump_table(%s): writing %d rows', table, row_count)             # <<<<<<<<<<<<<<
 *         write_integer(row_count, fp)
 * 
 */
          __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 356, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 356, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_row_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 356, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_1 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
              __pyx_t_10 = 1;
            }
          }
          __pyx_t_3 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 356, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (__pyx_t_1) {
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1); __pyx_t_1 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_dump_table_s_writing_d_rows);
          __Pyx_GIVEREF(__pyx_kp_u_dump_table_s_writing_d_rows);
          PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_10, __pyx_kp_u_dump_table_s_writing_d_rows);
          __Pyx_INCREF(__pyx_v_table);
          __Pyx_GIVEREF(__pyx_v_table);
          PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_10, __pyx_v_table);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_10, __pyx_t_4);
          __pyx_t_4 = 0;
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 356, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "s3ql/deltadump.pyx":357
 *         row_count = db.get_val("SELECT COUNT(rowid) FROM %s" % table)
 *         log.debug('dump_table(%s): writing %d rows', table, row_count)
 *         write_integer(row_count, fp)             # <<<<<<<<<<<<<<
 * 
 *         # Iterate through rows
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_write_integer(__pyx_v_row_count, __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 357, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":360
 * 
 *         # Iterate through rows
 *         while True:             # <<<<<<<<<<<<<<
 *             rc = sqlite3_step(stmt)
 *             if rc == SQLITE_DONE:
 */
          while (1) {

            /* "s3ql/deltadump.pyx":361
 *         # Iterate through rows
 *         while True:
 *             rc = sqlite3_step(stmt)             # <<<<<<<<<<<<<<
 *             if rc == SQLITE_DONE:
 *                 break
 */
            __pyx_v_rc = sqlite3_step(__pyx_cur_scope->__pyx_v_stmt);

            /* "s3ql/deltadump.pyx":362
 *         while True:
 *             rc = sqlite3_step(stmt)
 *             if rc == SQLITE_DONE:             # <<<<<<<<<<<<<<
 *                 break
 *             SQLITE_CHECK_RC(rc, SQLITE_ROW, sqlite3_db)
 */
            __pyx_t_2 = ((__pyx_v_rc == SQLITE_DONE) != 0);
            if (__pyx_t_2) {

              /* "s3ql/deltadump.pyx":363
 *             rc = sqlite3_step(stmt)
 *             if rc == SQLITE_DONE:
 *                 break             # <<<<<<<<<<<<<<
 *             SQLITE_CHECK_RC(rc, SQLITE_ROW, sqlite3_db)
 * 
 */
              goto __pyx_L23_break;

              /* "s3ql/deltadump.pyx":362
 *         while True:
 *             rc = sqlite3_step(stmt)
 *             if rc == SQLITE_DONE:             # <<<<<<<<<<<<<<
 *                 break
 *             SQLITE_CHECK_RC(rc, SQLITE_ROW, sqlite3_db)
 */
            }

            /* "s3ql/deltadump.pyx":364
 *             if rc == SQLITE_DONE:
 *                 break
 *             SQLITE_CHECK_RC(rc, SQLITE_ROW, sqlite3_db)             # <<<<<<<<<<<<<<
 * 
 *             for i in range(col_count):
 */
            __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(__pyx_v_rc, SQLITE_ROW, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 364, __pyx_L8_error)

            /* "s3ql/deltadump.pyx":366
 *             SQLITE_CHECK_RC(rc, SQLITE_ROW, sqlite3_db)
 * 
 *             for i in range(col_count):             # <<<<<<<<<<<<<<
 *                 if sqlite3_column_type(stmt, i) is SQLITE_NULL:
 *                     raise ValueError("Can't dump NULL values")
 */
            __pyx_t_13 = __pyx_v_col_count;
            for (__pyx_t_19 = 0; __pyx_t_19 < __pyx_t_13; __pyx_t_19+=1) {
              __pyx_v_i = __pyx_t_19;

              /* "s3ql/deltadump.pyx":367
 * 
 *             for i in range(col_count):
 *                 if sqlite3_column_type(stmt, i) is SQLITE_NULL:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Can't dump NULL values")
 * 
 */
              __pyx_t_2 = ((sqlite3_column_type(__pyx_cur_scope->__pyx_v_stmt, __pyx_v_i) == SQLITE_NULL) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":368
 *             for i in range(col_count):
 *                 if sqlite3_column_type(stmt, i) is SQLITE_NULL:
 *                     raise ValueError("Can't dump NULL values")             # <<<<<<<<<<<<<<
 * 
 *                 if col_types[i] == _INTEGER:
 */
                __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 368, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_Raise(__pyx_t_11, 0, 0, 0);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __PYX_ERR(0, 368, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":367
 * 
 *             for i in range(col_count):
 *                 if sqlite3_column_type(stmt, i) is SQLITE_NULL:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Can't dump NULL values")
 * 
 */
              }

              /* "s3ql/deltadump.pyx":370
 *                     raise ValueError("Can't dump NULL values")
 * 
 *                 if col_types[i] == _INTEGER:             # <<<<<<<<<<<<<<
 *                     int64 = sqlite3_column_int64(stmt, i)
 *                     tmp = int64
 */
              __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_types[__pyx_v_i]) == __pyx_v_4s3ql_9deltadump__INTEGER) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":371
 * 
 *                 if col_types[i] == _INTEGER:
 *                     int64 = sqlite3_column_int64(stmt, i)             # <<<<<<<<<<<<<<
 *                     tmp = int64
 *                     int64 -= int64_prev[i] + col_args[i]
 */
                __pyx_v_int64 = sqlite3_column_int64(__pyx_cur_scope->__pyx_v_stmt, __pyx_v_i);

                /* "s3ql/deltadump.pyx":372
 *                 if col_types[i] == _INTEGER:
 *                     int64 = sqlite3_column_int64(stmt, i)
 *                     tmp = int64             # <<<<<<<<<<<<<<
 *                     int64 -= int64_prev[i] + col_args[i]
 *                     int64_prev[i] = tmp
 */
                __pyx_v_tmp = __pyx_v_int64;

                /* "s3ql/deltadump.pyx":373
 *                     int64 = sqlite3_column_int64(stmt, i)
 *                     tmp = int64
 *                     int64 -= int64_prev[i] + col_args[i]             # <<<<<<<<<<<<<<
 *                     int64_prev[i] = tmp
 *                     write_integer(int64, fp)
 */
                __pyx_v_int64 = (__pyx_v_int64 - ((__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_i]) + (__pyx_cur_scope->__pyx_v_col_args[__pyx_v_i])));

                /* "s3ql/deltadump.pyx":374
 *                     tmp = int64
 *                     int64 -= int64_prev[i] + col_args[i]
 *                     int64_prev[i] = tmp             # <<<<<<<<<<<<<<
 *                     write_integer(int64, fp)
 * 
 */
                (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_i]) = __pyx_v_tmp;

                /* "s3ql/deltadump.pyx":375
 *                     int64 -= int64_prev[i] + col_args[i]
 *                     int64_prev[i] = tmp
 *                     write_integer(int64, fp)             # <<<<<<<<<<<<<<
 * 
 *                 elif col_types[i] == _TIME:
 */
                __pyx_t_20 = __pyx_f_4s3ql_9deltadump_write_integer(__pyx_v_int64, __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_20 == -1)) __PYX_ERR(0, 375, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":370
 *                     raise ValueError("Can't dump NULL values")
 * 
 *                 if col_types[i] == _INTEGER:             # <<<<<<<<<<<<<<
 *                     int64 = sqlite3_column_int64(stmt, i)
 *                     tmp = int64
 */
                goto __pyx_L28;
              }

              /* "s3ql/deltadump.pyx":377
 *                     write_integer(int64, fp)
 * 
 *                 elif col_types[i] == _TIME:             # <<<<<<<<<<<<<<
 *                     int64 = <int64_t> (sqlite3_column_double(stmt, i) * time_scale)
 *                     tmp = int64
 */
              __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_types[__pyx_v_i]) == __pyx_v_4s3ql_9deltadump__TIME) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":378
 * 
 *                 elif col_types[i] == _TIME:
 *                     int64 = <int64_t> (sqlite3_column_double(stmt, i) * time_scale)             # <<<<<<<<<<<<<<
 *                     tmp = int64
 *                     int64 -= int64_prev[i] + col_args[i]
 */
                __pyx_v_int64 = ((int64_t)(sqlite3_column_double(__pyx_cur_scope->__pyx_v_stmt, __pyx_v_i) * __pyx_v_4s3ql_9deltadump_time_scale));

                /* "s3ql/deltadump.pyx":379
 *                 elif col_types[i] == _TIME:
 *                     int64 = <int64_t> (sqlite3_column_double(stmt, i) * time_scale)
 *                     tmp = int64             # <<<<<<<<<<<<<<
 *                     int64 -= int64_prev[i] + col_args[i]
 *                     int64_prev[i] = tmp
 */
                __pyx_v_tmp = __pyx_v_int64;

                /* "s3ql/deltadump.pyx":380
 *                     int64 = <int64_t> (sqlite3_column_double(stmt, i) * time_scale)
 *                     tmp = int64
 *                     int64 -= int64_prev[i] + col_args[i]             # <<<<<<<<<<<<<<
 *                     int64_prev[i] = tmp
 *                     write_integer(int64, fp)
 */
                __pyx_v_int64 = (__pyx_v_int64 - ((__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_i]) + (__pyx_cur_scope->__pyx_v_col_args[__pyx_v_i])));

                /* "s3ql/deltadump.pyx":381
 *                     tmp = int64
 *                     int64 -= int64_prev[i] + col_args[i]
 *                     int64_prev[i] = tmp             # <<<<<<<<<<<<<<
 *                     write_integer(int64, fp)
 * 
 */
                (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_i]) = __pyx_v_tmp;

                /* "s3ql/deltadump.pyx":382
 *                     int64 -= int64_prev[i] + col_args[i]
 *                     int64_prev[i] = tmp
 *                     write_integer(int64, fp)             # <<<<<<<<<<<<<<
 * 
 *                 elif col_types[i] == _BLOB:
 */
                __pyx_t_20 = __pyx_f_4s3ql_9deltadump_write_integer(__pyx_v_int64, __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_20 == -1)) __PYX_ERR(0, 382, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":377
 *                     write_integer(int64, fp)
 * 
 *                 elif col_types[i] == _TIME:             # <<<<<<<<<<<<<<
 *                     int64 = <int64_t> (sqlite3_column_double(stmt, i) * time_scale)
 *                     tmp = int64
 */
                goto __pyx_L28;
              }

              /* "s3ql/deltadump.pyx":384
 *                     write_integer(int64, fp)
 * 
 *                 elif col_types[i] == _BLOB:             # <<<<<<<<<<<<<<
 *                     buf = sqlite3_column_blob(stmt, i)
 *                     rc = sqlite3_column_bytes(stmt, i)
 */
              __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_types[__pyx_v_i]) == __pyx_v_4s3ql_9deltadump__BLOB) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":385
 * 
 *                 elif col_types[i] == _BLOB:
 *                     buf = sqlite3_column_blob(stmt, i)             # <<<<<<<<<<<<<<
 *                     rc = sqlite3_column_bytes(stmt, i)
 *                     if rc > MAX_BLOB_SIZE:
 */
                __pyx_v_buf = sqlite3_column_blob(__pyx_cur_scope->__pyx_v_stmt, __pyx_v_i);

                /* "s3ql/deltadump.pyx":386
 *                 elif col_types[i] == _BLOB:
 *                     buf = sqlite3_column_blob(stmt, i)
 *                     rc = sqlite3_column_bytes(stmt, i)             # <<<<<<<<<<<<<<
 *                     if rc > MAX_BLOB_SIZE:
 *                             raise ValueError('Can not dump BLOB of size %d (max: %d)',
 */
                __pyx_v_rc = sqlite3_column_bytes(__pyx_cur_scope->__pyx_v_stmt, __pyx_v_i);

                /* "s3ql/deltadump.pyx":387
 *                     buf = sqlite3_column_blob(stmt, i)
 *                     rc = sqlite3_column_bytes(stmt, i)
 *                     if rc > MAX_BLOB_SIZE:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Can not dump BLOB of size %d (max: %d)',
 *                                              rc, MAX_BLOB_SIZE)
 */
                __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 387, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_11);
                __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_BLOB_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_6);
                __pyx_t_3 = PyObject_RichCompare(__pyx_t_11, __pyx_t_6, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L8_error)
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 387, __pyx_L8_error)
                __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":389
 *                     if rc > MAX_BLOB_SIZE:
 *                             raise ValueError('Can not dump BLOB of size %d (max: %d)',
 *                                              rc, MAX_BLOB_SIZE)             # <<<<<<<<<<<<<<
 *                     # Safe to cast now
 *                     len_ = <size_t> rc
 */
                  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_rc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 389, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_BLOB_SIZE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 389, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_6);

                  /* "s3ql/deltadump.pyx":388
 *                     rc = sqlite3_column_bytes(stmt, i)
 *                     if rc > MAX_BLOB_SIZE:
 *                             raise ValueError('Can not dump BLOB of size %d (max: %d)',             # <<<<<<<<<<<<<<
 *                                              rc, MAX_BLOB_SIZE)
 *                     # Safe to cast now
 */
                  __pyx_t_11 = PyTuple_New(3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_11);
                  __Pyx_INCREF(__pyx_kp_u_Can_not_dump_BLOB_of_size_d_max);
                  __Pyx_GIVEREF(__pyx_kp_u_Can_not_dump_BLOB_of_size_d_max);
                  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_kp_u_Can_not_dump_BLOB_of_size_d_max);
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_3);
                  __Pyx_GIVEREF(__pyx_t_6);
                  PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_t_6);
                  __pyx_t_3 = 0;
                  __pyx_t_6 = 0;
                  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 388, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                  __Pyx_Raise(__pyx_t_6, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
                  __PYX_ERR(0, 388, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":387
 *                     buf = sqlite3_column_blob(stmt, i)
 *                     rc = sqlite3_column_bytes(stmt, i)
 *                     if rc > MAX_BLOB_SIZE:             # <<<<<<<<<<<<<<
 *                             raise ValueError('Can not dump BLOB of size %d (max: %d)',
 *                                              rc, MAX_BLOB_SIZE)
 */
                }

                /* "s3ql/deltadump.pyx":391
 *                                              rc, MAX_BLOB_SIZE)
 *                     # Safe to cast now
 *                     len_ = <size_t> rc             # <<<<<<<<<<<<<<
 *                     if col_args[i] == 0:
 *                         write_integer(rc - int64_prev[i], fp)
 */
                __pyx_v_len_ = ((size_t)__pyx_v_rc);

                /* "s3ql/deltadump.pyx":392
 *                     # Safe to cast now
 *                     len_ = <size_t> rc
 *                     if col_args[i] == 0:             # <<<<<<<<<<<<<<
 *                         write_integer(rc - int64_prev[i], fp)
 *                         int64_prev[i] = rc
 */
                __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_args[__pyx_v_i]) == 0) != 0);
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":393
 *                     len_ = <size_t> rc
 *                     if col_args[i] == 0:
 *                         write_integer(rc - int64_prev[i], fp)             # <<<<<<<<<<<<<<
 *                         int64_prev[i] = rc
 *                     elif rc != col_args[i]:
 */
                  __pyx_t_20 = __pyx_f_4s3ql_9deltadump_write_integer((__pyx_v_rc - (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_i])), __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_20 == -1)) __PYX_ERR(0, 393, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":394
 *                     if col_args[i] == 0:
 *                         write_integer(rc - int64_prev[i], fp)
 *                         int64_prev[i] = rc             # <<<<<<<<<<<<<<
 *                     elif rc != col_args[i]:
 *                         raise ValueError("Length %d != %d in column %d" % (len_, col_args[i], i))
 */
                  (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_i]) = __pyx_v_rc;

                  /* "s3ql/deltadump.pyx":392
 *                     # Safe to cast now
 *                     len_ = <size_t> rc
 *                     if col_args[i] == 0:             # <<<<<<<<<<<<<<
 *                         write_integer(rc - int64_prev[i], fp)
 *                         int64_prev[i] = rc
 */
                  goto __pyx_L30;
                }

                /* "s3ql/deltadump.pyx":395
 *                         write_integer(rc - int64_prev[i], fp)
 *                         int64_prev[i] = rc
 *                     elif rc != col_args[i]:             # <<<<<<<<<<<<<<
 *                         raise ValueError("Length %d != %d in column %d" % (len_, col_args[i], i))
 * 
 */
                __pyx_t_2 = ((__pyx_v_rc != (__pyx_cur_scope->__pyx_v_col_args[__pyx_v_i])) != 0);
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":396
 *                         int64_prev[i] = rc
 *                     elif rc != col_args[i]:
 *                         raise ValueError("Length %d != %d in column %d" % (len_, col_args[i], i))             # <<<<<<<<<<<<<<
 * 
 *                     if len_ != 0:
 */
                  __pyx_t_6 = __Pyx_PyInt_FromSize_t(__pyx_v_len_); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_6);
                  __pyx_t_11 = __Pyx_PyInt_From_int((__pyx_cur_scope->__pyx_v_col_args[__pyx_v_i])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_11);
                  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_GIVEREF(__pyx_t_6);
                  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
                  __Pyx_GIVEREF(__pyx_t_11);
                  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_11);
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
                  __pyx_t_6 = 0;
                  __pyx_t_11 = 0;
                  __pyx_t_3 = 0;
                  __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_Length_d_d_in_column_d, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __Pyx_GIVEREF(__pyx_t_3);
                  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
                  __pyx_t_3 = 0;
                  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 396, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_3);
                  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                  __Pyx_Raise(__pyx_t_3, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
                  __PYX_ERR(0, 396, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":395
 *                         write_integer(rc - int64_prev[i], fp)
 *                         int64_prev[i] = rc
 *                     elif rc != col_args[i]:             # <<<<<<<<<<<<<<
 *                         raise ValueError("Length %d != %d in column %d" % (len_, col_args[i], i))
 * 
 */
                }
                __pyx_L30:;

                /* "s3ql/deltadump.pyx":398
 *                         raise ValueError("Length %d != %d in column %d" % (len_, col_args[i], i))
 * 
 *                     if len_ != 0:             # <<<<<<<<<<<<<<
 *                         fwrite(buf, len_, fp)
 * 
 */
                __pyx_t_2 = ((__pyx_v_len_ != 0) != 0);
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":399
 * 
 *                     if len_ != 0:
 *                         fwrite(buf, len_, fp)             # <<<<<<<<<<<<<<
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):
 */
                  __pyx_t_20 = __pyx_f_4s3ql_9deltadump_fwrite(__pyx_v_buf, __pyx_v_len_, __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_20 == -1)) __PYX_ERR(0, 399, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":398
 *                         raise ValueError("Length %d != %d in column %d" % (len_, col_args[i], i))
 * 
 *                     if len_ != 0:             # <<<<<<<<<<<<<<
 *                         fwrite(buf, len_, fp)
 * 
 */
                }

                /* "s3ql/deltadump.pyx":384
 *                     write_integer(int64, fp)
 * 
 *                 elif col_types[i] == _BLOB:             # <<<<<<<<<<<<<<
 *                     buf = sqlite3_column_blob(stmt, i)
 *                     rc = sqlite3_column_bytes(stmt, i)
 */
              }
              __pyx_L28:;
            }
          }
          __pyx_L23_break:;

          /* "s3ql/deltadump.pyx":321
 *         raise ValueError("Can't access in-memory databases")
 * 
 *     with ExitStack() as cm:             # <<<<<<<<<<<<<<
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L15_try_end;
        __pyx_L8_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("s3ql.deltadump.dump_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_4, &__pyx_t_11) < 0) __PYX_ERR(0, 321, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = PyTuple_Pack(3, __pyx_t_3, __pyx_t_4, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 321, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_21 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 321, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_21);
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_21);
          __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
          if (__pyx_t_2 < 0) __PYX_ERR(0, 321, __pyx_L10_except_error)
          __pyx_t_22 = ((!(__pyx_t_2 != 0)) != 0);
          if (__pyx_t_22) {
            __Pyx_GIVEREF(__pyx_t_3);
            __Pyx_GIVEREF(__pyx_t_4);
            __Pyx_XGIVEREF(__pyx_t_11);
            __Pyx_ErrRestoreWithState(__pyx_t_3, __pyx_t_4, __pyx_t_11);
            __pyx_t_3 = 0; __pyx_t_4 = 0; __pyx_t_11 = 0; 
            __PYX_ERR(0, 321, __pyx_L10_except_error)
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L9_exception_handled;
        }
        __pyx_L10_except_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L9_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L15_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_5) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__6, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 321, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L7;
      }
      __pyx_L7:;
    }
    goto __pyx_L35;
    __pyx_L4_error:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L1_error;
    __pyx_L35:;
  }

  /* "s3ql/deltadump.pyx":280
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 * def dump_table(table, order, columns, db, fh):             # <<<<<<<<<<<<<<
 *     '''Dump *columns* of *table* into *fh*
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("s3ql.deltadump.dump_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cm);
  __Pyx_XDECREF(__pyx_v_dbfile_b);
  __Pyx_XDECREF(__pyx_v_col_names);
  __Pyx_XDECREF(__pyx_v_query);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":401
 *                         fwrite(buf, len_, fp)
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):             # <<<<<<<<<<<<<<
 *     '''Load *columns* of *table* from *fh*
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_5load_table(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4s3ql_9deltadump_4load_table[] = "load_table(table, columns, db, fh, trx_rows=5000)\nLoad *columns* of *table* from *fh*\n\n    *db* is an `s3ql.Connection` instance for the database.\n\n    *columns* must be the same list of 3-tuples that was passed to\n    `dump_table` when creating the dump stored in *fh*.\n\n    This function will open a separate connection to the database, so\n    the *db* connection should not be in EXCLUSIVE locking mode.\n    (Using a separate connection avoids the requirement on the *apsw*\n    and *deltadump* modules be linked against against binary\n    compatible SQLite libraries).\n\n    When writing into the table, a new transaction will be started\n    every *trx_rows* rows.\n    ";
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_5load_table = {"load_table", (PyCFunction)__pyx_pw_4s3ql_9deltadump_5load_table, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4s3ql_9deltadump_4load_table};
static PyObject *__pyx_pw_4s3ql_9deltadump_5load_table(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_table = 0;
  PyObject *__pyx_v_columns = 0;
  PyObject *__pyx_v_db = 0;
  PyObject *__pyx_v_fh = 0;
  PyObject *__pyx_v_trx_rows = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("load_table (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_table,&__pyx_n_s_columns,&__pyx_n_s_db,&__pyx_n_s_fh,&__pyx_n_s_trx_rows,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[4] = ((PyObject *)__pyx_int_5000);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_table)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_columns)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("load_table", 0, 4, 5, 1); __PYX_ERR(0, 401, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_db)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("load_table", 0, 4, 5, 2); __PYX_ERR(0, 401, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_fh)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("load_table", 0, 4, 5, 3); __PYX_ERR(0, 401, __pyx_L3_error)
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_trx_rows);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "load_table") < 0)) __PYX_ERR(0, 401, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_table = values[0];
    __pyx_v_columns = values[1];
    __pyx_v_db = values[2];
    __pyx_v_fh = values[3];
    __pyx_v_trx_rows = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("load_table", 0, 4, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 401, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("s3ql.deltadump.load_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4s3ql_9deltadump_4load_table(__pyx_self, __pyx_v_table, __pyx_v_columns, __pyx_v_db, __pyx_v_fh, __pyx_v_trx_rows);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":441
 *                                         SQLITE_OPEN_READWRITE, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_lambda6(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_lambda6 = {"lambda6", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_lambda6, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_lambda6(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda6 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda6(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda6(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda6", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":442
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),
 *                                             SQLITE_OK, sqlite3_db))             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 *                         SQLITE_OK, sqlite3_db)
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_close(__pyx_cur_scope->__pyx_v_sqlite3_db), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 441, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":441
 *                                         SQLITE_OPEN_READWRITE, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda6", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":462
 *         # Get FILE* for buffered reading from *fh*
 *         fp = dup_to_fp(fh, b'rb')
 *         cm.callback(lambda: fclose(fp))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate col_args and col_types
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_1lambda7(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_1lambda7 = {"lambda7", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_1lambda7, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_1lambda7(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda7 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda7(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda7(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda7", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_fclose(__pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 462, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda7", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":466
 *         # Allocate col_args and col_types
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(col_types))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_2lambda8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_2lambda8 = {"lambda8", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_2lambda8, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_2lambda8(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda8 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda8(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda8(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda8", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_col_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":467
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))
 *         cm.callback(lambda: free(col_types))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate int64_prev
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_3lambda9(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_3lambda9 = {"lambda9", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_3lambda9, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_3lambda9(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda9 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda9(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda9(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda9", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_col_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda9", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":471
 *         # Allocate int64_prev
 *         int64_prev = <int64_t*> calloc(<size_t> len(columns), sizeof(int64_t))
 *         cm.callback(lambda: free(int64_prev))             # <<<<<<<<<<<<<<
 * 
 *         # Prepare INSERT statement
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_4lambda10(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_4lambda10 = {"lambda10", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_4lambda10, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_4lambda10(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda10 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda10(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda10(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda10", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_int64_prev); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda10", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":480
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_5lambda11(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_5lambda11 = {"lambda11", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_5lambda11, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_5lambda11(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda11 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda11(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda11(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda11", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":481
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),
 *                                                  SQLITE_OK, sqlite3_db))             # <<<<<<<<<<<<<<
 * 
 *         # Prepare BEGIN statement
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_finalize(__pyx_cur_scope->__pyx_v_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 480, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":480
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda11", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":487
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &begin_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(begin_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_6lambda12(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_6lambda12 = {"lambda12", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_6lambda12, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_6lambda12(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda12 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda12(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda12(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda12", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":488
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(begin_stmt),
 *                                                  SQLITE_OK, sqlite3_db))             # <<<<<<<<<<<<<<
 * 
 *         # Prepare COMMIT statement
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_finalize(__pyx_cur_scope->__pyx_v_begin_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 487, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":487
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &begin_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(begin_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda12", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":494
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &commit_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(commit_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_7lambda13(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_7lambda13 = {"lambda13", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_7lambda13, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_7lambda13(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda13 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda13(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda13(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda13", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":495
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(commit_stmt),
 *                                                  SQLITE_OK, sqlite3_db))             # <<<<<<<<<<<<<<
 * 
 *         buf = calloc(MAX_BLOB_SIZE, 1)
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_finalize(__pyx_cur_scope->__pyx_v_commit_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 494, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":494
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &commit_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(commit_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda13", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":498
 * 
 *         buf = calloc(MAX_BLOB_SIZE, 1)
 *         cm.callback(lambda: free(buf))             # <<<<<<<<<<<<<<
 *         read_integer(&row_count, fp)
 *         log.debug('load_table(%s): reading %d rows', table, row_count)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_8lambda14(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_8lambda14 = {"lambda14", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_8lambda14, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_8lambda14(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda14 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda14(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda14(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("lambda14", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_free(__pyx_cur_scope->__pyx_v_buf); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda14", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":504
 *         # Start transaction
 *         SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_step(commit_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_DONE, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 */

/* Python wrapper */
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_9lambda15(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_4s3ql_9deltadump_10load_table_9lambda15 = {"lambda15", (PyCFunction)__pyx_pw_4s3ql_9deltadump_10load_table_9lambda15, METH_NOARGS, 0};
static PyObject *__pyx_pw_4s3ql_9deltadump_10load_table_9lambda15(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda15 (wrapper)", 0);
  __pyx_r = __pyx_lambda_funcdef_lambda15(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda15(PyObject *__pyx_self) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("lambda15", 0);
  __pyx_outer_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);

  /* "s3ql/deltadump.pyx":505
 *         SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_step(commit_stmt),
 *                                                  SQLITE_DONE, sqlite3_db))             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 * 
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_step(__pyx_cur_scope->__pyx_v_commit_stmt), SQLITE_DONE, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 504, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":504
 *         # Start transaction
 *         SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_step(commit_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_DONE, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("s3ql.deltadump.load_table.lambda15", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":401
 *                         fwrite(buf, len_, fp)
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):             # <<<<<<<<<<<<<<
 *     '''Load *columns* of *table* from *fh*
 * 
 */

static PyObject *__pyx_pf_4s3ql_9deltadump_4load_table(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_table, PyObject *__pyx_v_columns, PyObject *__pyx_v_db, PyObject *__pyx_v_fh, PyObject *__pyx_v_trx_rows) {
  struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_cur_scope;
  int __pyx_v_col_count;
  int __pyx_v_rc;
  int __pyx_v_len_;
  int __pyx_v_i;
  int __pyx_v_j;
  int64_t __pyx_v_row_count;
  int64_t __pyx_v_int64;
  int64_t __pyx_v_tmp;
  PyObject *__pyx_v_cm = NULL;
  PyObject *__pyx_v_dbfile_b = NULL;
  PyObject *__pyx_v_pragma = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_cmd = NULL;
  PyObject *__pyx_v_col_names = NULL;
  PyObject *__pyx_v_query = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  const char *__pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  char *__pyx_t_15;
  int __pyx_t_16;
  char const *__pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_t_24;
  FILE *__pyx_t_25;
  void *__pyx_t_26;
  PyObject *(*__pyx_t_27)(PyObject *);
  size_t __pyx_t_28;
  int64_t __pyx_t_29;
  int __pyx_t_30;
  int __pyx_t_31;
  __Pyx_RefNannySetupContext("load_table", 0);
  __pyx_cur_scope = (struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *)__pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct_1_load_table(__pyx_ptype_4s3ql_9deltadump___pyx_scope_struct_1_load_table, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);

  /* "s3ql/deltadump.pyx":431
 *     cdef int64_t row_count, int64, tmp
 * 
 *     if db.file == ':memory:':             # <<<<<<<<<<<<<<
 *         raise ValueError("Can't access in-memory databases")
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_kp_u_memory, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 431, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":432
 * 
 *     if db.file == ':memory:':
 *         raise ValueError("Can't access in-memory databases")             # <<<<<<<<<<<<<<
 * 
 *     with ExitStack() as cm:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 432, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":431
 *     cdef int64_t row_count, int64, tmp
 * 
 *     if db.file == ':memory:':             # <<<<<<<<<<<<<<
 *         raise ValueError("Can't access in-memory databases")
 * 
 */
  }

  /* "s3ql/deltadump.pyx":434
 *         raise ValueError("Can't access in-memory databases")
 * 
 *     with ExitStack() as cm:             # <<<<<<<<<<<<<<
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 */
  /*with:*/ {
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ExitStack); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (__pyx_t_4) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_exit); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 434, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_PyObject_LookupSpecial(__pyx_t_1, __pyx_n_s_enter); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L4_error)
    }
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        /*try:*/ {
          __pyx_v_cm = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":436
 *     with ExitStack() as cm:
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)             # <<<<<<<<<<<<<<
 *         dbfile_b = db.file.encode(sys.getfilesystemencoding(), 'surrogateescape')
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 436, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_file); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_10 = 1;
            }
          }
          __pyx_t_11 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 436, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_Opening_connection_to_s);
          __Pyx_GIVEREF(__pyx_kp_u_Opening_connection_to_s);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_10, __pyx_kp_u_Opening_connection_to_s);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_10, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_11, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 436, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":437
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 *         dbfile_b = db.file.encode(sys.getfilesystemencoding(), 'surrogateescape')             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,
 *                                         SQLITE_OPEN_READWRITE, NULL),
 */
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_file); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 437, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_sys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getfilesystemencoding); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 437, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          if (__pyx_t_1) {
            __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L8_error)
          }
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_10 = 1;
            }
          }
          __pyx_t_1 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (__pyx_t_6) {
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_3);
          PyTuple_SET_ITEM(__pyx_t_1, 0+__pyx_t_10, __pyx_t_3);
          __Pyx_INCREF(__pyx_n_u_surrogateescape);
          __Pyx_GIVEREF(__pyx_n_u_surrogateescape);
          PyTuple_SET_ITEM(__pyx_t_1, 1+__pyx_t_10, __pyx_n_u_surrogateescape);
          __pyx_t_3 = 0;
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 437, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_dbfile_b = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":438
 *         log.debug('Opening connection to %s', db.file)
 *         dbfile_b = db.file.encode(sys.getfilesystemencoding(), 'surrogateescape')
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,             # <<<<<<<<<<<<<<
 *                                         SQLITE_OPEN_READWRITE, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __pyx_t_12 = __Pyx_PyObject_AsString(__pyx_v_dbfile_b); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 438, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":440
 *         SQLITE_CHECK_RC(sqlite3_open_v2(dbfile_b, &sqlite3_db,
 *                                         SQLITE_OPEN_READWRITE, NULL),
 *                         SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),
 *                                             SQLITE_OK, sqlite3_db))
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_open_v2(__pyx_t_12, (&__pyx_cur_scope->__pyx_v_sqlite3_db), SQLITE_OPEN_READWRITE, NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 438, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":441
 *                                         SQLITE_OPEN_READWRITE, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),             # <<<<<<<<<<<<<<
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 441, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_lambda6, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_3) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 441, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 441, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_3); __pyx_t_3 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 441, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":443
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_close(sqlite3_db),
 *                                             SQLITE_OK, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_extended_result_codes(sqlite3_db, 1),             # <<<<<<<<<<<<<<
 *                         SQLITE_OK, sqlite3_db)
 * 
 */
          __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_extended_result_codes(__pyx_cur_scope->__pyx_v_sqlite3_db, 1), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 443, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":447
 * 
 *         # Copy settings
 *         for pragma in ('synchronous', 'foreign_keys'):             # <<<<<<<<<<<<<<
 *             val = db.get_val('PRAGMA %s' % pragma)
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')
 */
          __pyx_t_4 = __pyx_tuple__8; __Pyx_INCREF(__pyx_t_4); __pyx_t_10 = 0;
          for (;;) {
            if (__pyx_t_10 >= 2) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_11); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 447, __pyx_L8_error)
            #else
            __pyx_t_11 = PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 447, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_11);
            #endif
            __Pyx_XDECREF_SET(__pyx_v_pragma, ((PyObject*)__pyx_t_11));
            __pyx_t_11 = 0;

            /* "s3ql/deltadump.pyx":448
 *         # Copy settings
 *         for pragma in ('synchronous', 'foreign_keys'):
 *             val = db.get_val('PRAGMA %s' % pragma)             # <<<<<<<<<<<<<<
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')
 *             SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, cmd, -1, &stmt, NULL),
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_db, __pyx_n_s_get_val); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 448, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_PRAGMA_s, __pyx_v_pragma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = NULL;
            if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            if (!__pyx_t_3) {
              __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 448, __pyx_L8_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_11);
            } else {
              __pyx_t_14 = PyTuple_New(1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 448, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_3); __pyx_t_3 = NULL;
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_14, 0+1, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_14, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 448, __pyx_L8_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            }
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "s3ql/deltadump.pyx":449
 *         for pragma in ('synchronous', 'foreign_keys'):
 *             val = db.get_val('PRAGMA %s' % pragma)
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')             # <<<<<<<<<<<<<<
 *             SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, cmd, -1, &stmt, NULL),
 *                             SQLITE_OK, sqlite3_db)
 */
            __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 449, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_INCREF(__pyx_v_pragma);
            __Pyx_GIVEREF(__pyx_v_pragma);
            PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_v_pragma);
            __Pyx_INCREF(__pyx_v_val);
            __Pyx_GIVEREF(__pyx_v_val);
            PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_v_val);
            __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_PRAGMA_s_s, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 449, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = PyUnicode_AsUTF8String(((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 449, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_XDECREF_SET(__pyx_v_cmd, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "s3ql/deltadump.pyx":450
 *             val = db.get_val('PRAGMA %s' % pragma)
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')
 *             SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, cmd, -1, &stmt, NULL),             # <<<<<<<<<<<<<<
 *                             SQLITE_OK, sqlite3_db)
 *             try:
 */
            __pyx_t_15 = __Pyx_PyObject_AsString(__pyx_v_cmd); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 450, __pyx_L8_error)

            /* "s3ql/deltadump.pyx":451
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')
 *             SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, cmd, -1, &stmt, NULL),
 *                             SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *             try:
 *                 rc = sqlite3_step(stmt)
 */
            __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_prepare_v2(__pyx_cur_scope->__pyx_v_sqlite3_db, __pyx_t_15, -1, (&__pyx_cur_scope->__pyx_v_stmt), NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 450, __pyx_L8_error)

            /* "s3ql/deltadump.pyx":452
 *             SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, cmd, -1, &stmt, NULL),
 *                             SQLITE_OK, sqlite3_db)
 *             try:             # <<<<<<<<<<<<<<
 *                 rc = sqlite3_step(stmt)
 *                 if rc == SQLITE_ROW:
 */
            /*try:*/ {

              /* "s3ql/deltadump.pyx":453
 *                             SQLITE_OK, sqlite3_db)
 *             try:
 *                 rc = sqlite3_step(stmt)             # <<<<<<<<<<<<<<
 *                 if rc == SQLITE_ROW:
 *                     rc = sqlite3_step(stmt)
 */
              __pyx_v_rc = sqlite3_step(__pyx_cur_scope->__pyx_v_stmt);

              /* "s3ql/deltadump.pyx":454
 *             try:
 *                 rc = sqlite3_step(stmt)
 *                 if rc == SQLITE_ROW:             # <<<<<<<<<<<<<<
 *                     rc = sqlite3_step(stmt)
 *                 SQLITE_CHECK_RC(rc, SQLITE_DONE, sqlite3_db)
 */
              __pyx_t_2 = ((__pyx_v_rc == SQLITE_ROW) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":455
 *                 rc = sqlite3_step(stmt)
 *                 if rc == SQLITE_ROW:
 *                     rc = sqlite3_step(stmt)             # <<<<<<<<<<<<<<
 *                 SQLITE_CHECK_RC(rc, SQLITE_DONE, sqlite3_db)
 *             finally:
 */
                __pyx_v_rc = sqlite3_step(__pyx_cur_scope->__pyx_v_stmt);

                /* "s3ql/deltadump.pyx":454
 *             try:
 *                 rc = sqlite3_step(stmt)
 *                 if rc == SQLITE_ROW:             # <<<<<<<<<<<<<<
 *                     rc = sqlite3_step(stmt)
 *                 SQLITE_CHECK_RC(rc, SQLITE_DONE, sqlite3_db)
 */
              }

              /* "s3ql/deltadump.pyx":456
 *                 if rc == SQLITE_ROW:
 *                     rc = sqlite3_step(stmt)
 *                 SQLITE_CHECK_RC(rc, SQLITE_DONE, sqlite3_db)             # <<<<<<<<<<<<<<
 *             finally:
 *                 SQLITE_CHECK_RC(sqlite3_finalize(stmt), SQLITE_OK, sqlite3_db)
 */
              __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(__pyx_v_rc, SQLITE_DONE, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 456, __pyx_L21_error)
            }

            /* "s3ql/deltadump.pyx":458
 *                 SQLITE_CHECK_RC(rc, SQLITE_DONE, sqlite3_db)
 *             finally:
 *                 SQLITE_CHECK_RC(sqlite3_finalize(stmt), SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 * 
 *         # Get FILE* for buffered reading from *fh*
 */
            /*finally:*/ {
              /*normal exit:*/{
                __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_finalize(__pyx_cur_scope->__pyx_v_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 458, __pyx_L8_error)
                goto __pyx_L22;
              }
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __pyx_L21_error:;
                __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
                __Pyx_PyThreadState_assign
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
                __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
                __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20) < 0)) __Pyx_ErrFetch(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_18);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_21);
                __Pyx_XGOTREF(__pyx_t_22);
                __Pyx_XGOTREF(__pyx_t_23);
                __pyx_t_13 = __pyx_lineno; __pyx_t_16 = __pyx_clineno; __pyx_t_17 = __pyx_filename;
                {
                  __pyx_t_24 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_finalize(__pyx_cur_scope->__pyx_v_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_24 == -1)) __PYX_ERR(0, 458, __pyx_L27_error)
                }
                __Pyx_PyThreadState_assign
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_21);
                  __Pyx_XGIVEREF(__pyx_t_22);
                  __Pyx_XGIVEREF(__pyx_t_23);
                  __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
                }
                __Pyx_XGIVEREF(__pyx_t_18);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ErrRestore(__pyx_t_18, __pyx_t_19, __pyx_t_20);
                __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
                __pyx_lineno = __pyx_t_13; __pyx_clineno = __pyx_t_16; __pyx_filename = __pyx_t_17;
                goto __pyx_L8_error;
                __pyx_L27_error:;
                __Pyx_PyThreadState_assign
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_21);
                  __Pyx_XGIVEREF(__pyx_t_22);
                  __Pyx_XGIVEREF(__pyx_t_23);
                  __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_22, __pyx_t_23);
                }
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
                __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
                __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0;
                goto __pyx_L8_error;
              }
              __pyx_L22:;
            }

            /* "s3ql/deltadump.pyx":447
 * 
 *         # Copy settings
 *         for pragma in ('synchronous', 'foreign_keys'):             # <<<<<<<<<<<<<<
 *             val = db.get_val('PRAGMA %s' % pragma)
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')
 */
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":461
 * 
 *         # Get FILE* for buffered reading from *fh*
 *         fp = dup_to_fp(fh, b'rb')             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: fclose(fp))
 * 
 */
          __pyx_t_25 = __pyx_f_4s3ql_9deltadump_dup_to_fp(__pyx_v_fh, ((const char *)"rb")); if (unlikely(__pyx_t_25 == NULL)) __PYX_ERR(0, 461, __pyx_L8_error)
          __pyx_cur_scope->__pyx_v_fp = __pyx_t_25;

          /* "s3ql/deltadump.pyx":462
 *         # Get FILE* for buffered reading from *fh*
 *         fp = dup_to_fp(fh, b'rb')
 *         cm.callback(lambda: fclose(fp))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate col_args and col_types
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 462, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_1lambda7, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 462, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_14) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 462, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_14); __pyx_t_14 = NULL;
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 462, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":465
 * 
 *         # Allocate col_args and col_types
 *         col_count = prep_columns(columns, &col_types, &col_args)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(col_args))
 *         cm.callback(lambda: free(col_types))
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_prep_columns(__pyx_v_columns, (&__pyx_cur_scope->__pyx_v_col_types), (&__pyx_cur_scope->__pyx_v_col_args)); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 465, __pyx_L8_error)
          __pyx_v_col_count = __pyx_t_16;

          /* "s3ql/deltadump.pyx":466
 *         # Allocate col_args and col_types
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(col_types))
 * 
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 466, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_2lambda8, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_6) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 466, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_14 = PyTuple_New(1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 466, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_GIVEREF(__pyx_t_1);
            PyTuple_SET_ITEM(__pyx_t_14, 0+1, __pyx_t_1);
            __pyx_t_1 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 466, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":467
 *         col_count = prep_columns(columns, &col_types, &col_args)
 *         cm.callback(lambda: free(col_args))
 *         cm.callback(lambda: free(col_types))             # <<<<<<<<<<<<<<
 * 
 *         # Allocate int64_prev
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 467, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_14 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_3lambda9, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 467, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_1 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_1) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_14); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 467, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 467, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
            __Pyx_GIVEREF(__pyx_t_14);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_14);
            __pyx_t_14 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 467, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":470
 * 
 *         # Allocate int64_prev
 *         int64_prev = <int64_t*> calloc(<size_t> len(columns), sizeof(int64_t))             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(int64_prev))
 * 
 */
          __pyx_t_10 = PyObject_Length(__pyx_v_columns); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 470, __pyx_L8_error)
          __pyx_t_26 = __pyx_f_4s3ql_9deltadump_calloc(((size_t)__pyx_t_10), (sizeof(int64_t))); if (unlikely(__pyx_t_26 == NULL)) __PYX_ERR(0, 470, __pyx_L8_error)
          __pyx_cur_scope->__pyx_v_int64_prev = ((int64_t *)__pyx_t_26);

          /* "s3ql/deltadump.pyx":471
 *         # Allocate int64_prev
 *         int64_prev = <int64_t*> calloc(<size_t> len(columns), sizeof(int64_t))
 *         cm.callback(lambda: free(int64_prev))             # <<<<<<<<<<<<<<
 * 
 *         # Prepare INSERT statement
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 471, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_4lambda10, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 471, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_14) {
            __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 471, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_4);
          } else {
            __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 471, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_14); __pyx_t_14 = NULL;
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_1, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 471, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":474
 * 
 *         # Prepare INSERT statement
 *         col_names = [ x[0] for x in columns ]             # <<<<<<<<<<<<<<
 *         query = ("INSERT INTO %s (%s) VALUES(%s)"
 *                  % (table, ', '.join(col_names),
 */
          { /* enter inner scope */
            PyObject *__pyx_8genexpr1__pyx_v_x = NULL;
            __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 474, __pyx_L30_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (likely(PyList_CheckExact(__pyx_v_columns)) || PyTuple_CheckExact(__pyx_v_columns)) {
              __pyx_t_11 = __pyx_v_columns; __Pyx_INCREF(__pyx_t_11); __pyx_t_10 = 0;
              __pyx_t_27 = NULL;
            } else {
              __pyx_t_10 = -1; __pyx_t_11 = PyObject_GetIter(__pyx_v_columns); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 474, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_27 = Py_TYPE(__pyx_t_11)->tp_iternext; if (unlikely(!__pyx_t_27)) __PYX_ERR(0, 474, __pyx_L30_error)
            }
            for (;;) {
              if (likely(!__pyx_t_27)) {
                if (likely(PyList_CheckExact(__pyx_t_11))) {
                  if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_11)) break;
                  #if CYTHON_COMPILING_IN_CPYTHON
                  __pyx_t_1 = PyList_GET_ITEM(__pyx_t_11, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 474, __pyx_L30_error)
                  #else
                  __pyx_t_1 = PySequence_ITEM(__pyx_t_11, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L30_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  #endif
                } else {
                  if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_11)) break;
                  #if CYTHON_COMPILING_IN_CPYTHON
                  __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_11, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 474, __pyx_L30_error)
                  #else
                  __pyx_t_1 = PySequence_ITEM(__pyx_t_11, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L30_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  #endif
                }
              } else {
                __pyx_t_1 = __pyx_t_27(__pyx_t_11);
                if (unlikely(!__pyx_t_1)) {
                  PyObject* exc_type = PyErr_Occurred();
                  if (exc_type) {
                    if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                    else __PYX_ERR(0, 474, __pyx_L30_error)
                  }
                  break;
                }
                __Pyx_GOTREF(__pyx_t_1);
              }
              __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_x, __pyx_t_1);
              __pyx_t_1 = 0;
              __pyx_t_1 = __Pyx_GetItemInt(__pyx_8genexpr1__pyx_v_x, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L30_error)
              __Pyx_GOTREF(__pyx_t_1);
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 474, __pyx_L30_error)
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            }
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x);
            goto __pyx_L33_exit_scope;
            __pyx_L30_error:;
            __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_x);
            goto __pyx_L8_error;
            __pyx_L33_exit_scope:;
          } /* exit inner scope */
          __pyx_v_col_names = ((PyObject*)__pyx_t_4);
          __pyx_t_4 = 0;

          /* "s3ql/deltadump.pyx":476
 *         col_names = [ x[0] for x in columns ]
 *         query = ("INSERT INTO %s (%s) VALUES(%s)"
 *                  % (table, ', '.join(col_names),             # <<<<<<<<<<<<<<
 *                     ', '.join('?' * col_count))).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 */
          __pyx_t_4 = PyUnicode_Join(__pyx_kp_u__4, __pyx_v_col_names); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 476, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);

          /* "s3ql/deltadump.pyx":477
 *         query = ("INSERT INTO %s (%s) VALUES(%s)"
 *                  % (table, ', '.join(col_names),
 *                     ', '.join('?' * col_count))).encode('utf-8')             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_col_count); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 477, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_1 = PyNumber_Multiply(__pyx_kp_u__9, __pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = PyUnicode_Join(__pyx_kp_u__4, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 477, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":476
 *         col_names = [ x[0] for x in columns ]
 *         query = ("INSERT INTO %s (%s) VALUES(%s)"
 *                  % (table, ', '.join(col_names),             # <<<<<<<<<<<<<<
 *                     ', '.join('?' * col_count))).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 */
          __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_v_table);
          __Pyx_GIVEREF(__pyx_v_table);
          PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_table);
          __Pyx_GIVEREF(__pyx_t_4);
          PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_11);
          __pyx_t_4 = 0;
          __pyx_t_11 = 0;
          __pyx_t_11 = PyUnicode_Format(__pyx_kp_u_INSERT_INTO_s_s_VALUES_s, __pyx_t_1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 476, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":477
 *         query = ("INSERT INTO %s (%s) VALUES(%s)"
 *                  % (table, ', '.join(col_names),
 *                     ', '.join('?' * col_count))).encode('utf-8')             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __pyx_t_1 = PyUnicode_AsUTF8String(((PyObject*)__pyx_t_11)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 477, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_v_query = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":478
 *                  % (table, ', '.join(col_names),
 *                     ', '.join('?' * col_count))).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),             # <<<<<<<<<<<<<<
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),
 */
          __pyx_t_15 = __Pyx_PyObject_AsString(__pyx_v_query); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":479
 *                     ', '.join('?' * col_count))).encode('utf-8')
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),
 *                                                  SQLITE_OK, sqlite3_db))
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_prepare_v2(__pyx_cur_scope->__pyx_v_sqlite3_db, __pyx_t_15, -1, (&__pyx_cur_scope->__pyx_v_stmt), NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 478, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":480
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 480, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_5lambda11, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_6) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_14 = PyTuple_New(1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 480, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_14, 0+1, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":484
 * 
 *         # Prepare BEGIN statement
 *         query = b'BEGIN TRANSACTION'             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &begin_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __Pyx_INCREF(__pyx_kp_b_BEGIN_TRANSACTION);
          __Pyx_DECREF_SET(__pyx_v_query, __pyx_kp_b_BEGIN_TRANSACTION);

          /* "s3ql/deltadump.pyx":485
 *         # Prepare BEGIN statement
 *         query = b'BEGIN TRANSACTION'
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &begin_stmt, NULL),             # <<<<<<<<<<<<<<
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(begin_stmt),
 */
          __pyx_t_15 = __Pyx_PyObject_AsString(__pyx_v_query); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":486
 *         query = b'BEGIN TRANSACTION'
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &begin_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(begin_stmt),
 *                                                  SQLITE_OK, sqlite3_db))
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_prepare_v2(__pyx_cur_scope->__pyx_v_sqlite3_db, __pyx_t_15, -1, (&__pyx_cur_scope->__pyx_v_begin_stmt), NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 485, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":487
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &begin_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(begin_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 487, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_14 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_6lambda12, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 487, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_4 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_4) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_14); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 487, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
            __Pyx_GIVEREF(__pyx_t_14);
            PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_14);
            __pyx_t_14 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":491
 * 
 *         # Prepare COMMIT statement
 *         query = b'COMMIT TRANSACTION'             # <<<<<<<<<<<<<<
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &commit_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 */
          __Pyx_INCREF(__pyx_kp_b_COMMIT_TRANSACTION);
          __Pyx_DECREF_SET(__pyx_v_query, __pyx_kp_b_COMMIT_TRANSACTION);

          /* "s3ql/deltadump.pyx":492
 *         # Prepare COMMIT statement
 *         query = b'COMMIT TRANSACTION'
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &commit_stmt, NULL),             # <<<<<<<<<<<<<<
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(commit_stmt),
 */
          __pyx_t_15 = __Pyx_PyObject_AsString(__pyx_v_query); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":493
 *         query = b'COMMIT TRANSACTION'
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &commit_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(commit_stmt),
 *                                                  SQLITE_OK, sqlite3_db))
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_prepare_v2(__pyx_cur_scope->__pyx_v_sqlite3_db, __pyx_t_15, -1, (&__pyx_cur_scope->__pyx_v_commit_stmt), NULL), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 492, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":494
 *         SQLITE_CHECK_RC(sqlite3_prepare_v2(sqlite3_db, query, -1, &commit_stmt, NULL),
 *                         SQLITE_OK, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_finalize(commit_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 494, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_7lambda13, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 494, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_14) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_14); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_14); __pyx_t_14 = NULL;
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":497
 *                                                  SQLITE_OK, sqlite3_db))
 * 
 *         buf = calloc(MAX_BLOB_SIZE, 1)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: free(buf))
 *         read_integer(&row_count, fp)
 */
          __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_BLOB_SIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 497, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_28 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_28 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 497, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_26 = __pyx_f_4s3ql_9deltadump_calloc(__pyx_t_28, 1); if (unlikely(__pyx_t_26 == NULL)) __PYX_ERR(0, 497, __pyx_L8_error)
          __pyx_cur_scope->__pyx_v_buf = __pyx_t_26;

          /* "s3ql/deltadump.pyx":498
 * 
 *         buf = calloc(MAX_BLOB_SIZE, 1)
 *         cm.callback(lambda: free(buf))             # <<<<<<<<<<<<<<
 *         read_integer(&row_count, fp)
 *         log.debug('load_table(%s): reading %d rows', table, row_count)
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 498, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_8lambda14, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 498, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_6)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_6);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          if (!__pyx_t_6) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_14 = PyTuple_New(1+1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 498, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_6); __pyx_t_6 = NULL;
            __Pyx_GIVEREF(__pyx_t_4);
            PyTuple_SET_ITEM(__pyx_t_14, 0+1, __pyx_t_4);
            __pyx_t_4 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 498, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":499
 *         buf = calloc(MAX_BLOB_SIZE, 1)
 *         cm.callback(lambda: free(buf))
 *         read_integer(&row_count, fp)             # <<<<<<<<<<<<<<
 *         log.debug('load_table(%s): reading %d rows', table, row_count)
 * 
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_read_integer((&__pyx_v_row_count), __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 499, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":500
 *         cm.callback(lambda: free(buf))
 *         read_integer(&row_count, fp)
 *         log.debug('load_table(%s): reading %d rows', table, row_count)             # <<<<<<<<<<<<<<
 * 
 *         # Start transaction
 */
          __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 500, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 500, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyInt_From_int64_t(__pyx_v_row_count); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 500, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_4 = NULL;
          __pyx_t_10 = 0;
          if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
              __pyx_t_10 = 1;
            }
          }
          __pyx_t_6 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 500, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_load_table_s_reading_d_rows);
          __Pyx_GIVEREF(__pyx_kp_u_load_table_s_reading_d_rows);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_kp_u_load_table_s_reading_d_rows);
          __Pyx_INCREF(__pyx_v_table);
          __Pyx_GIVEREF(__pyx_v_table);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_v_table);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_10, __pyx_t_11);
          __pyx_t_11 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":503
 * 
 *         # Start transaction
 *         SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)             # <<<<<<<<<<<<<<
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_step(commit_stmt),
 *                                                  SQLITE_DONE, sqlite3_db))
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_step(__pyx_cur_scope->__pyx_v_begin_stmt), SQLITE_DONE, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 503, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":504
 *         # Start transaction
 *         SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_step(commit_stmt),             # <<<<<<<<<<<<<<
 *                                                  SQLITE_DONE, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_cm, __pyx_n_s_callback); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 504, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_6 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_10load_table_9lambda15, 0, __pyx_n_s_load_table_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_s3ql_deltadump, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 504, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_11 = NULL;
          if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
            }
          }
          if (!__pyx_t_11) {
            __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else {
            __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 504, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_11); __pyx_t_11 = NULL;
            __Pyx_GIVEREF(__pyx_t_6);
            PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "s3ql/deltadump.pyx":506
 *         cm.callback(lambda: SQLITE_CHECK_RC(sqlite3_step(commit_stmt),
 *                                                  SQLITE_DONE, sqlite3_db))
 *         SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 * 
 *         # Iterate through rows
 */
          __pyx_t_16 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_reset(__pyx_cur_scope->__pyx_v_begin_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_16 == -1)) __PYX_ERR(0, 506, __pyx_L8_error)

          /* "s3ql/deltadump.pyx":509
 * 
 *         # Iterate through rows
 *         for i in range(row_count):             # <<<<<<<<<<<<<<
 *             for j in range(col_count):
 *                 if col_types[j] == _INTEGER:
 */
          __pyx_t_29 = __pyx_v_row_count;
          for (__pyx_t_16 = 0; __pyx_t_16 < __pyx_t_29; __pyx_t_16+=1) {
            __pyx_v_i = __pyx_t_16;

            /* "s3ql/deltadump.pyx":510
 *         # Iterate through rows
 *         for i in range(row_count):
 *             for j in range(col_count):             # <<<<<<<<<<<<<<
 *                 if col_types[j] == _INTEGER:
 *                     read_integer(&int64, fp)
 */
            __pyx_t_13 = __pyx_v_col_count;
            for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_13; __pyx_t_24+=1) {
              __pyx_v_j = __pyx_t_24;

              /* "s3ql/deltadump.pyx":511
 *         for i in range(row_count):
 *             for j in range(col_count):
 *                 if col_types[j] == _INTEGER:             # <<<<<<<<<<<<<<
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]
 */
              __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_types[__pyx_v_j]) == __pyx_v_4s3ql_9deltadump__INTEGER) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":512
 *             for j in range(col_count):
 *                 if col_types[j] == _INTEGER:
 *                     read_integer(&int64, fp)             # <<<<<<<<<<<<<<
 *                     int64 += col_args[j] + int64_prev[j]
 *                     int64_prev[j] = int64
 */
                __pyx_t_30 = __pyx_f_4s3ql_9deltadump_read_integer((&__pyx_v_int64), __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 512, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":513
 *                 if col_types[j] == _INTEGER:
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]             # <<<<<<<<<<<<<<
 *                     int64_prev[j] = int64
 *                     SQLITE_CHECK_RC(sqlite3_bind_int64(stmt, j + 1, int64),
 */
                __pyx_v_int64 = (__pyx_v_int64 + ((__pyx_cur_scope->__pyx_v_col_args[__pyx_v_j]) + (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_j])));

                /* "s3ql/deltadump.pyx":514
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]
 *                     int64_prev[j] = int64             # <<<<<<<<<<<<<<
 *                     SQLITE_CHECK_RC(sqlite3_bind_int64(stmt, j + 1, int64),
 *                                     SQLITE_OK, sqlite3_db)
 */
                (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_j]) = __pyx_v_int64;

                /* "s3ql/deltadump.pyx":515
 *                     int64 += col_args[j] + int64_prev[j]
 *                     int64_prev[j] = int64
 *                     SQLITE_CHECK_RC(sqlite3_bind_int64(stmt, j + 1, int64),             # <<<<<<<<<<<<<<
 *                                     SQLITE_OK, sqlite3_db)
 * 
 */
                __pyx_t_30 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_bind_int64(__pyx_cur_scope->__pyx_v_stmt, (__pyx_v_j + 1), __pyx_v_int64), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 515, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":511
 *         for i in range(row_count):
 *             for j in range(col_count):
 *                 if col_types[j] == _INTEGER:             # <<<<<<<<<<<<<<
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]
 */
              }

              /* "s3ql/deltadump.pyx":518
 *                                     SQLITE_OK, sqlite3_db)
 * 
 *                 if col_types[j] == _TIME:             # <<<<<<<<<<<<<<
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]
 */
              __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_types[__pyx_v_j]) == __pyx_v_4s3ql_9deltadump__TIME) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":519
 * 
 *                 if col_types[j] == _TIME:
 *                     read_integer(&int64, fp)             # <<<<<<<<<<<<<<
 *                     int64 += col_args[j] + int64_prev[j]
 *                     int64_prev[j] = int64
 */
                __pyx_t_30 = __pyx_f_4s3ql_9deltadump_read_integer((&__pyx_v_int64), __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 519, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":520
 *                 if col_types[j] == _TIME:
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]             # <<<<<<<<<<<<<<
 *                     int64_prev[j] = int64
 *                     # Cast is safe, we know that the integer was converted from
 */
                __pyx_v_int64 = (__pyx_v_int64 + ((__pyx_cur_scope->__pyx_v_col_args[__pyx_v_j]) + (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_j])));

                /* "s3ql/deltadump.pyx":521
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]
 *                     int64_prev[j] = int64             # <<<<<<<<<<<<<<
 *                     # Cast is safe, we know that the integer was converted from
 *                     # double at dump time.
 */
                (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_j]) = __pyx_v_int64;

                /* "s3ql/deltadump.pyx":524
 *                     # Cast is safe, we know that the integer was converted from
 *                     # double at dump time.
 *                     SQLITE_CHECK_RC(sqlite3_bind_double(stmt, j + 1, <double> int64 / time_scale),             # <<<<<<<<<<<<<<
 *                                     SQLITE_OK, sqlite3_db)
 * 
 */
                if (unlikely(__pyx_v_4s3ql_9deltadump_time_scale == 0)) {
                  PyErr_SetString(PyExc_ZeroDivisionError, "float division");
                  __PYX_ERR(0, 524, __pyx_L8_error)
                }

                /* "s3ql/deltadump.pyx":525
 *                     # double at dump time.
 *                     SQLITE_CHECK_RC(sqlite3_bind_double(stmt, j + 1, <double> int64 / time_scale),
 *                                     SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 * 
 *                 elif col_types[j] == _BLOB:
 */
                __pyx_t_30 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_bind_double(__pyx_cur_scope->__pyx_v_stmt, (__pyx_v_j + 1), (((double)__pyx_v_int64) / __pyx_v_4s3ql_9deltadump_time_scale)), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 524, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":518
 *                                     SQLITE_OK, sqlite3_db)
 * 
 *                 if col_types[j] == _TIME:             # <<<<<<<<<<<<<<
 *                     read_integer(&int64, fp)
 *                     int64 += col_args[j] + int64_prev[j]
 */
                goto __pyx_L39;
              }

              /* "s3ql/deltadump.pyx":527
 *                                     SQLITE_OK, sqlite3_db)
 * 
 *                 elif col_types[j] == _BLOB:             # <<<<<<<<<<<<<<
 *                     if col_args[j] == 0:
 *                         read_integer(&int64, fp)
 */
              __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_types[__pyx_v_j]) == __pyx_v_4s3ql_9deltadump__BLOB) != 0);
              if (__pyx_t_2) {

                /* "s3ql/deltadump.pyx":528
 * 
 *                 elif col_types[j] == _BLOB:
 *                     if col_args[j] == 0:             # <<<<<<<<<<<<<<
 *                         read_integer(&int64, fp)
 *                         tmp = int64_prev[j] + int64
 */
                __pyx_t_2 = (((__pyx_cur_scope->__pyx_v_col_args[__pyx_v_j]) == 0) != 0);
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":529
 *                 elif col_types[j] == _BLOB:
 *                     if col_args[j] == 0:
 *                         read_integer(&int64, fp)             # <<<<<<<<<<<<<<
 *                         tmp = int64_prev[j] + int64
 *                         if tmp < 0 or tmp > INT_MAX:
 */
                  __pyx_t_30 = __pyx_f_4s3ql_9deltadump_read_integer((&__pyx_v_int64), __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 529, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":530
 *                     if col_args[j] == 0:
 *                         read_integer(&int64, fp)
 *                         tmp = int64_prev[j] + int64             # <<<<<<<<<<<<<<
 *                         if tmp < 0 or tmp > INT_MAX:
 *                             raise RuntimeError('Corrupted input')
 */
                  __pyx_v_tmp = ((__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_j]) + __pyx_v_int64);

                  /* "s3ql/deltadump.pyx":531
 *                         read_integer(&int64, fp)
 *                         tmp = int64_prev[j] + int64
 *                         if tmp < 0 or tmp > INT_MAX:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError('Corrupted input')
 *                         len_ = <int> tmp
 */
                  __pyx_t_31 = ((__pyx_v_tmp < 0) != 0);
                  if (!__pyx_t_31) {
                  } else {
                    __pyx_t_2 = __pyx_t_31;
                    goto __pyx_L42_bool_binop_done;
                  }
                  __pyx_t_31 = ((__pyx_v_tmp > INT_MAX) != 0);
                  __pyx_t_2 = __pyx_t_31;
                  __pyx_L42_bool_binop_done:;
                  if (__pyx_t_2) {

                    /* "s3ql/deltadump.pyx":532
 *                         tmp = int64_prev[j] + int64
 *                         if tmp < 0 or tmp > INT_MAX:
 *                             raise RuntimeError('Corrupted input')             # <<<<<<<<<<<<<<
 *                         len_ = <int> tmp
 *                         int64_prev[j] = tmp
 */
                    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 532, __pyx_L8_error)
                    __Pyx_GOTREF(__pyx_t_1);
                    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
                    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                    __PYX_ERR(0, 532, __pyx_L8_error)

                    /* "s3ql/deltadump.pyx":531
 *                         read_integer(&int64, fp)
 *                         tmp = int64_prev[j] + int64
 *                         if tmp < 0 or tmp > INT_MAX:             # <<<<<<<<<<<<<<
 *                             raise RuntimeError('Corrupted input')
 *                         len_ = <int> tmp
 */
                  }

                  /* "s3ql/deltadump.pyx":533
 *                         if tmp < 0 or tmp > INT_MAX:
 *                             raise RuntimeError('Corrupted input')
 *                         len_ = <int> tmp             # <<<<<<<<<<<<<<
 *                         int64_prev[j] = tmp
 *                     else:
 */
                  __pyx_v_len_ = ((int)__pyx_v_tmp);

                  /* "s3ql/deltadump.pyx":534
 *                             raise RuntimeError('Corrupted input')
 *                         len_ = <int> tmp
 *                         int64_prev[j] = tmp             # <<<<<<<<<<<<<<
 *                     else:
 *                         len_ = col_args[j]
 */
                  (__pyx_cur_scope->__pyx_v_int64_prev[__pyx_v_j]) = __pyx_v_tmp;

                  /* "s3ql/deltadump.pyx":528
 * 
 *                 elif col_types[j] == _BLOB:
 *                     if col_args[j] == 0:             # <<<<<<<<<<<<<<
 *                         read_integer(&int64, fp)
 *                         tmp = int64_prev[j] + int64
 */
                  goto __pyx_L40;
                }

                /* "s3ql/deltadump.pyx":536
 *                         int64_prev[j] = tmp
 *                     else:
 *                         len_ = col_args[j]             # <<<<<<<<<<<<<<
 * 
 *                     if len_ > MAX_BLOB_SIZE:
 */
                /*else*/ {
                  __pyx_v_len_ = (__pyx_cur_scope->__pyx_v_col_args[__pyx_v_j]);
                }
                __pyx_L40:;

                /* "s3ql/deltadump.pyx":538
 *                         len_ = col_args[j]
 * 
 *                     if len_ > MAX_BLOB_SIZE:             # <<<<<<<<<<<<<<
 *                         raise RuntimeError('BLOB too large to read (%d vs %d)', len_, MAX_BLOB_SIZE)
 * 
 */
                __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_len_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_1);
                __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_BLOB_SIZE); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 538, __pyx_L8_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_14, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 538, __pyx_L8_error)
                __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 538, __pyx_L8_error)
                __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":539
 * 
 *                     if len_ > MAX_BLOB_SIZE:
 *                         raise RuntimeError('BLOB too large to read (%d vs %d)', len_, MAX_BLOB_SIZE)             # <<<<<<<<<<<<<<
 * 
 *                     if len_ > 0:
 */
                  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_len_); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 539, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_4);
                  __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_MAX_BLOB_SIZE); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 539, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_14);
                  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_1);
                  __Pyx_INCREF(__pyx_kp_u_BLOB_too_large_to_read_d_vs_d);
                  __Pyx_GIVEREF(__pyx_kp_u_BLOB_too_large_to_read_d_vs_d);
                  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_BLOB_too_large_to_read_d_vs_d);
                  __Pyx_GIVEREF(__pyx_t_4);
                  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
                  __Pyx_GIVEREF(__pyx_t_14);
                  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_14);
                  __pyx_t_4 = 0;
                  __pyx_t_14 = 0;
                  __pyx_t_14 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_t_1, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 539, __pyx_L8_error)
                  __Pyx_GOTREF(__pyx_t_14);
                  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
                  __Pyx_Raise(__pyx_t_14, 0, 0, 0);
                  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                  __PYX_ERR(0, 539, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":538
 *                         len_ = col_args[j]
 * 
 *                     if len_ > MAX_BLOB_SIZE:             # <<<<<<<<<<<<<<
 *                         raise RuntimeError('BLOB too large to read (%d vs %d)', len_, MAX_BLOB_SIZE)
 * 
 */
                }

                /* "s3ql/deltadump.pyx":541
 *                         raise RuntimeError('BLOB too large to read (%d vs %d)', len_, MAX_BLOB_SIZE)
 * 
 *                     if len_ > 0:             # <<<<<<<<<<<<<<
 *                         fread(buf, <unsigned> len_, fp)
 * 
 */
                __pyx_t_2 = ((__pyx_v_len_ > 0) != 0);
                if (__pyx_t_2) {

                  /* "s3ql/deltadump.pyx":542
 * 
 *                     if len_ > 0:
 *                         fread(buf, <unsigned> len_, fp)             # <<<<<<<<<<<<<<
 * 
 *                     SQLITE_CHECK_RC(sqlite3_bind_blob(stmt, j + 1, buf, len_, SQLITE_TRANSIENT),
 */
                  __pyx_t_30 = __pyx_f_4s3ql_9deltadump_fread(__pyx_cur_scope->__pyx_v_buf, ((unsigned int)__pyx_v_len_), __pyx_cur_scope->__pyx_v_fp); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 542, __pyx_L8_error)

                  /* "s3ql/deltadump.pyx":541
 *                         raise RuntimeError('BLOB too large to read (%d vs %d)', len_, MAX_BLOB_SIZE)
 * 
 *                     if len_ > 0:             # <<<<<<<<<<<<<<
 *                         fread(buf, <unsigned> len_, fp)
 * 
 */
                }

                /* "s3ql/deltadump.pyx":544
 *                         fread(buf, <unsigned> len_, fp)
 * 
 *                     SQLITE_CHECK_RC(sqlite3_bind_blob(stmt, j + 1, buf, len_, SQLITE_TRANSIENT),             # <<<<<<<<<<<<<<
 *                                     SQLITE_OK, sqlite3_db)
 * 
 */
                __pyx_t_30 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_bind_blob(__pyx_cur_scope->__pyx_v_stmt, (__pyx_v_j + 1), __pyx_cur_scope->__pyx_v_buf, __pyx_v_len_, SQLITE_TRANSIENT), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_30 == -1)) __PYX_ERR(0, 544, __pyx_L8_error)

                /* "s3ql/deltadump.pyx":527
 *                                     SQLITE_OK, sqlite3_db)
 * 
 *                 elif col_types[j] == _BLOB:             # <<<<<<<<<<<<<<
 *                     if col_args[j] == 0:
 *                         read_integer(&int64, fp)
 */
              }
              __pyx_L39:;
            }

            /* "s3ql/deltadump.pyx":547
 *                                     SQLITE_OK, sqlite3_db)
 * 
 *             SQLITE_CHECK_RC(sqlite3_step(stmt), SQLITE_DONE, sqlite3_db)             # <<<<<<<<<<<<<<
 *             SQLITE_CHECK_RC(sqlite3_reset(stmt), SQLITE_OK, sqlite3_db)
 * 
 */
            __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_step(__pyx_cur_scope->__pyx_v_stmt), SQLITE_DONE, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 547, __pyx_L8_error)

            /* "s3ql/deltadump.pyx":548
 * 
 *             SQLITE_CHECK_RC(sqlite3_step(stmt), SQLITE_DONE, sqlite3_db)
 *             SQLITE_CHECK_RC(sqlite3_reset(stmt), SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 * 
 *             # Commit every once in a while
 */
            __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_reset(__pyx_cur_scope->__pyx_v_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 548, __pyx_L8_error)

            /* "s3ql/deltadump.pyx":551
 * 
 *             # Commit every once in a while
 *             if i % trx_rows == 0:             # <<<<<<<<<<<<<<
 *                 # This isn't 100% ok -- if we have an exception in step(begin_stmt),
 *                 # we the cleanup handler will execute the commit statement again
 */
            __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 551, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_1 = PyNumber_Remainder(__pyx_t_14, __pyx_v_trx_rows); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_14 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 551, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 551, __pyx_L8_error)
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (__pyx_t_2) {

              /* "s3ql/deltadump.pyx":555
 *                 # we the cleanup handler will execute the commit statement again
 *                 # without an active transaction.
 *                 SQLITE_CHECK_RC(sqlite3_step(commit_stmt), SQLITE_DONE, sqlite3_db)             # <<<<<<<<<<<<<<
 *                 SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_reset(commit_stmt), SQLITE_OK, sqlite3_db)
 */
              __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_step(__pyx_cur_scope->__pyx_v_commit_stmt), SQLITE_DONE, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 555, __pyx_L8_error)

              /* "s3ql/deltadump.pyx":556
 *                 # without an active transaction.
 *                 SQLITE_CHECK_RC(sqlite3_step(commit_stmt), SQLITE_DONE, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)             # <<<<<<<<<<<<<<
 *                 SQLITE_CHECK_RC(sqlite3_reset(commit_stmt), SQLITE_OK, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 */
              __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_step(__pyx_cur_scope->__pyx_v_begin_stmt), SQLITE_DONE, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 556, __pyx_L8_error)

              /* "s3ql/deltadump.pyx":557
 *                 SQLITE_CHECK_RC(sqlite3_step(commit_stmt), SQLITE_DONE, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_reset(commit_stmt), SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 *                 SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 * 
 */
              __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_reset(__pyx_cur_scope->__pyx_v_commit_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 557, __pyx_L8_error)

              /* "s3ql/deltadump.pyx":558
 *                 SQLITE_CHECK_RC(sqlite3_step(begin_stmt), SQLITE_DONE, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_reset(commit_stmt), SQLITE_OK, sqlite3_db)
 *                 SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)             # <<<<<<<<<<<<<<
 * 
 * cdef inline int write_integer(int64_t int64, FILE * fp) except -1:
 */
              __pyx_t_13 = __pyx_f_4s3ql_9deltadump_SQLITE_CHECK_RC(sqlite3_reset(__pyx_cur_scope->__pyx_v_begin_stmt), SQLITE_OK, __pyx_cur_scope->__pyx_v_sqlite3_db); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 558, __pyx_L8_error)

              /* "s3ql/deltadump.pyx":551
 * 
 *             # Commit every once in a while
 *             if i % trx_rows == 0:             # <<<<<<<<<<<<<<
 *                 # This isn't 100% ok -- if we have an exception in step(begin_stmt),
 *                 # we the cleanup handler will execute the commit statement again
 */
            }
          }

          /* "s3ql/deltadump.pyx":434
 *         raise ValueError("Can't access in-memory databases")
 * 
 *     with ExitStack() as cm:             # <<<<<<<<<<<<<<
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 */
        }
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L15_try_end;
        __pyx_L8_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        /*except:*/ {
          __Pyx_AddTraceback("s3ql.deltadump.load_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_14, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(0, 434, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_6 = PyTuple_Pack(3, __pyx_t_14, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 434, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_23 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 434, __pyx_L10_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
          if (__pyx_t_2 < 0) __PYX_ERR(0, 434, __pyx_L10_except_error)
          __pyx_t_31 = ((!(__pyx_t_2 != 0)) != 0);
          if (__pyx_t_31) {
            __Pyx_GIVEREF(__pyx_t_14);
            __Pyx_GIVEREF(__pyx_t_1);
            __Pyx_XGIVEREF(__pyx_t_4);
            __Pyx_ErrRestoreWithState(__pyx_t_14, __pyx_t_1, __pyx_t_4);
            __pyx_t_14 = 0; __pyx_t_1 = 0; __pyx_t_4 = 0; 
            __PYX_ERR(0, 434, __pyx_L10_except_error)
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L9_exception_handled;
        }
        __pyx_L10_except_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        goto __pyx_L1_error;
        __pyx_L9_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
        __pyx_L15_try_end:;
      }
    }
    /*finally:*/ {
      /*normal exit:*/{
        if (__pyx_t_5) {
          __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__11, NULL);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 434, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        goto __pyx_L7;
      }
      __pyx_L7:;
    }
    goto __pyx_L50;
    __pyx_L4_error:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L1_error;
    __pyx_L50:;
  }

  /* "s3ql/deltadump.pyx":401
 *                         fwrite(buf, len_, fp)
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):             # <<<<<<<<<<<<<<
 *     '''Load *columns* of *table* from *fh*
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("s3ql.deltadump.load_table", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cm);
  __Pyx_XDECREF(__pyx_v_dbfile_b);
  __Pyx_XDECREF(__pyx_v_pragma);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_cmd);
  __Pyx_XDECREF(__pyx_v_col_names);
  __Pyx_XDECREF(__pyx_v_query);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":560
 *                 SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 * 
 * cdef inline int write_integer(int64_t int64, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Write *int64* into *fp*, using as little space as possible
 * 
 */

static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_write_integer(int64_t __pyx_v_int64, FILE *__pyx_v_fp) {
  unsigned int __pyx_v_int8;
  uint8_t __pyx_v_int8_real;
  size_t __pyx_v_len_;
  uint64_t __pyx_v_uint64;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  uint64_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("write_integer", 0);

  /* "s3ql/deltadump.pyx":576
 *     cdef uint64_t uint64
 * 
 *     if int64 < 0:             # <<<<<<<<<<<<<<
 *         uint64 = <uint64_t> -int64
 *         int8 = 0x80 # Highest bit set
 */
  __pyx_t_1 = ((__pyx_v_int64 < 0) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":577
 * 
 *     if int64 < 0:
 *         uint64 = <uint64_t> -int64             # <<<<<<<<<<<<<<
 *         int8 = 0x80 # Highest bit set
 *     else:
 */
    __pyx_v_uint64 = ((uint64_t)(-__pyx_v_int64));

    /* "s3ql/deltadump.pyx":578
 *     if int64 < 0:
 *         uint64 = <uint64_t> -int64
 *         int8 = 0x80 # Highest bit set             # <<<<<<<<<<<<<<
 *     else:
 *         uint64 = <uint64_t> int64
 */
    __pyx_v_int8 = 0x80;

    /* "s3ql/deltadump.pyx":576
 *     cdef uint64_t uint64
 * 
 *     if int64 < 0:             # <<<<<<<<<<<<<<
 *         uint64 = <uint64_t> -int64
 *         int8 = 0x80 # Highest bit set
 */
    goto __pyx_L3;
  }

  /* "s3ql/deltadump.pyx":580
 *         int8 = 0x80 # Highest bit set
 *     else:
 *         uint64 = <uint64_t> int64             # <<<<<<<<<<<<<<
 *         int8 = 0
 * 
 */
  /*else*/ {
    __pyx_v_uint64 = ((uint64_t)__pyx_v_int64);

    /* "s3ql/deltadump.pyx":581
 *     else:
 *         uint64 = <uint64_t> int64
 *         int8 = 0             # <<<<<<<<<<<<<<
 * 
 *     if uint64 < 0x80 and uint64 not in (INT8, INT16, INT32, INT64):
 */
    __pyx_v_int8 = 0;
  }
  __pyx_L3:;

  /* "s3ql/deltadump.pyx":583
 *         int8 = 0
 * 
 *     if uint64 < 0x80 and uint64 not in (INT8, INT16, INT32, INT64):             # <<<<<<<<<<<<<<
 *         len_ = 0
 *         int8 += <uint8_t> uint64
 */
  __pyx_t_2 = ((__pyx_v_uint64 < 0x80) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __pyx_v_uint64;
  __pyx_t_4 = ((__pyx_t_3 != __pyx_v_4s3ql_9deltadump_INT8) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_t_3 != __pyx_v_4s3ql_9deltadump_INT16) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_t_3 != __pyx_v_4s3ql_9deltadump_INT32) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  __pyx_t_4 = ((__pyx_t_3 != __pyx_v_4s3ql_9deltadump_INT64) != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L7_bool_binop_done:;
  __pyx_t_4 = (__pyx_t_2 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":584
 * 
 *     if uint64 < 0x80 and uint64 not in (INT8, INT16, INT32, INT64):
 *         len_ = 0             # <<<<<<<<<<<<<<
 *         int8 += <uint8_t> uint64
 *     elif uint64 < UINT8_MAX:
 */
    __pyx_v_len_ = 0;

    /* "s3ql/deltadump.pyx":585
 *     if uint64 < 0x80 and uint64 not in (INT8, INT16, INT32, INT64):
 *         len_ = 0
 *         int8 += <uint8_t> uint64             # <<<<<<<<<<<<<<
 *     elif uint64 < UINT8_MAX:
 *         len_ = 1
 */
    __pyx_v_int8 = (__pyx_v_int8 + ((uint8_t)__pyx_v_uint64));

    /* "s3ql/deltadump.pyx":583
 *         int8 = 0
 * 
 *     if uint64 < 0x80 and uint64 not in (INT8, INT16, INT32, INT64):             # <<<<<<<<<<<<<<
 *         len_ = 0
 *         int8 += <uint8_t> uint64
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":586
 *         len_ = 0
 *         int8 += <uint8_t> uint64
 *     elif uint64 < UINT8_MAX:             # <<<<<<<<<<<<<<
 *         len_ = 1
 *         int8 += INT8
 */
  __pyx_t_1 = ((__pyx_v_uint64 < UINT8_MAX) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":587
 *         int8 += <uint8_t> uint64
 *     elif uint64 < UINT8_MAX:
 *         len_ = 1             # <<<<<<<<<<<<<<
 *         int8 += INT8
 *     elif uint64 < UINT16_MAX:
 */
    __pyx_v_len_ = 1;

    /* "s3ql/deltadump.pyx":588
 *     elif uint64 < UINT8_MAX:
 *         len_ = 1
 *         int8 += INT8             # <<<<<<<<<<<<<<
 *     elif uint64 < UINT16_MAX:
 *         len_ = 2
 */
    __pyx_v_int8 = (__pyx_v_int8 + __pyx_v_4s3ql_9deltadump_INT8);

    /* "s3ql/deltadump.pyx":586
 *         len_ = 0
 *         int8 += <uint8_t> uint64
 *     elif uint64 < UINT8_MAX:             # <<<<<<<<<<<<<<
 *         len_ = 1
 *         int8 += INT8
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":589
 *         len_ = 1
 *         int8 += INT8
 *     elif uint64 < UINT16_MAX:             # <<<<<<<<<<<<<<
 *         len_ = 2
 *         int8 += INT16
 */
  __pyx_t_1 = ((__pyx_v_uint64 < UINT16_MAX) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":590
 *         int8 += INT8
 *     elif uint64 < UINT16_MAX:
 *         len_ = 2             # <<<<<<<<<<<<<<
 *         int8 += INT16
 *     elif uint64 < UINT32_MAX:
 */
    __pyx_v_len_ = 2;

    /* "s3ql/deltadump.pyx":591
 *     elif uint64 < UINT16_MAX:
 *         len_ = 2
 *         int8 += INT16             # <<<<<<<<<<<<<<
 *     elif uint64 < UINT32_MAX:
 *         len_ = 4
 */
    __pyx_v_int8 = (__pyx_v_int8 + __pyx_v_4s3ql_9deltadump_INT16);

    /* "s3ql/deltadump.pyx":589
 *         len_ = 1
 *         int8 += INT8
 *     elif uint64 < UINT16_MAX:             # <<<<<<<<<<<<<<
 *         len_ = 2
 *         int8 += INT16
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":592
 *         len_ = 2
 *         int8 += INT16
 *     elif uint64 < UINT32_MAX:             # <<<<<<<<<<<<<<
 *         len_ = 4
 *         int8 += INT32
 */
  __pyx_t_1 = ((__pyx_v_uint64 < UINT32_MAX) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":593
 *         int8 += INT16
 *     elif uint64 < UINT32_MAX:
 *         len_ = 4             # <<<<<<<<<<<<<<
 *         int8 += INT32
 *     else:
 */
    __pyx_v_len_ = 4;

    /* "s3ql/deltadump.pyx":594
 *     elif uint64 < UINT32_MAX:
 *         len_ = 4
 *         int8 += INT32             # <<<<<<<<<<<<<<
 *     else:
 *         len_ = 8
 */
    __pyx_v_int8 = (__pyx_v_int8 + __pyx_v_4s3ql_9deltadump_INT32);

    /* "s3ql/deltadump.pyx":592
 *         len_ = 2
 *         int8 += INT16
 *     elif uint64 < UINT32_MAX:             # <<<<<<<<<<<<<<
 *         len_ = 4
 *         int8 += INT32
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":596
 *         int8 += INT32
 *     else:
 *         len_ = 8             # <<<<<<<<<<<<<<
 *         int8 += INT64
 * 
 */
  /*else*/ {
    __pyx_v_len_ = 8;

    /* "s3ql/deltadump.pyx":597
 *     else:
 *         len_ = 8
 *         int8 += INT64             # <<<<<<<<<<<<<<
 * 
 *     # Cast
 */
    __pyx_v_int8 = (__pyx_v_int8 + __pyx_v_4s3ql_9deltadump_INT64);
  }
  __pyx_L4:;

  /* "s3ql/deltadump.pyx":600
 * 
 *     # Cast
 *     int8_real = <uint8_t> int8             # <<<<<<<<<<<<<<
 *     fwrite(&int8_real, 1, fp)
 *     if len_ != 0:
 */
  __pyx_v_int8_real = ((uint8_t)__pyx_v_int8);

  /* "s3ql/deltadump.pyx":601
 *     # Cast
 *     int8_real = <uint8_t> int8
 *     fwrite(&int8_real, 1, fp)             # <<<<<<<<<<<<<<
 *     if len_ != 0:
 *         uint64 = htole64(uint64)
 */
  __pyx_t_5 = __pyx_f_4s3ql_9deltadump_fwrite((&__pyx_v_int8_real), 1, __pyx_v_fp); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 601, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":602
 *     int8_real = <uint8_t> int8
 *     fwrite(&int8_real, 1, fp)
 *     if len_ != 0:             # <<<<<<<<<<<<<<
 *         uint64 = htole64(uint64)
 *         fwrite(&uint64, len_, fp)
 */
  __pyx_t_1 = ((__pyx_v_len_ != 0) != 0);
  if (__pyx_t_1) {

    /* "s3ql/deltadump.pyx":603
 *     fwrite(&int8_real, 1, fp)
 *     if len_ != 0:
 *         uint64 = htole64(uint64)             # <<<<<<<<<<<<<<
 *         fwrite(&uint64, len_, fp)
 * 
 */
    __pyx_v_uint64 = htole64(__pyx_v_uint64);

    /* "s3ql/deltadump.pyx":604
 *     if len_ != 0:
 *         uint64 = htole64(uint64)
 *         fwrite(&uint64, len_, fp)             # <<<<<<<<<<<<<<
 * 
 *     # len <= 8, safe to cast
 */
    __pyx_t_5 = __pyx_f_4s3ql_9deltadump_fwrite((&__pyx_v_uint64), __pyx_v_len_, __pyx_v_fp); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 604, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":602
 *     int8_real = <uint8_t> int8
 *     fwrite(&int8_real, 1, fp)
 *     if len_ != 0:             # <<<<<<<<<<<<<<
 *         uint64 = htole64(uint64)
 *         fwrite(&uint64, len_, fp)
 */
  }

  /* "s3ql/deltadump.pyx":607
 * 
 *     # len <= 8, safe to cast
 *     return <int> len_ + 1             # <<<<<<<<<<<<<<
 * 
 * cdef inline int read_integer(int64_t * out, FILE * fp) except -1:
 */
  __pyx_r = (((int)__pyx_v_len_) + 1);
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":560
 *                 SQLITE_CHECK_RC(sqlite3_reset(begin_stmt), SQLITE_OK, sqlite3_db)
 * 
 * cdef inline int write_integer(int64_t int64, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Write *int64* into *fp*, using as little space as possible
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("s3ql.deltadump.write_integer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "s3ql/deltadump.pyx":609
 *     return <int> len_ + 1
 * 
 * cdef inline int read_integer(int64_t * out, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Read integer written using `write_integer` from *fp*
 * 
 */

static CYTHON_INLINE int __pyx_f_4s3ql_9deltadump_read_integer(int64_t *__pyx_v_out, FILE *__pyx_v_fp) {
  uint8_t __pyx_v_int8;
  size_t __pyx_v_len_;
  uint64_t __pyx_v_uint64;
  char __pyx_v_negative;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("read_integer", 0);

  /* "s3ql/deltadump.pyx":620
 *     cdef char negative
 * 
 *     fread(&int8, 1, fp)             # <<<<<<<<<<<<<<
 * 
 *     if int8 & 0x80 != 0:
 */
  __pyx_t_1 = __pyx_f_4s3ql_9deltadump_fread((&__pyx_v_int8), 1, __pyx_v_fp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 620, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":622
 *     fread(&int8, 1, fp)
 * 
 *     if int8 & 0x80 != 0:             # <<<<<<<<<<<<<<
 *         negative = 1
 *         # Need to cast again due to integer promotion
 */
  __pyx_t_2 = (((__pyx_v_int8 & 0x80) != 0) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":623
 * 
 *     if int8 & 0x80 != 0:
 *         negative = 1             # <<<<<<<<<<<<<<
 *         # Need to cast again due to integer promotion
 *         int8 = <uint8_t> (int8 & (~ 0x80))
 */
    __pyx_v_negative = 1;

    /* "s3ql/deltadump.pyx":625
 *         negative = 1
 *         # Need to cast again due to integer promotion
 *         int8 = <uint8_t> (int8 & (~ 0x80))             # <<<<<<<<<<<<<<
 *     else:
 *         negative = 0
 */
    __pyx_v_int8 = ((uint8_t)(__pyx_v_int8 & (~0x80)));

    /* "s3ql/deltadump.pyx":622
 *     fread(&int8, 1, fp)
 * 
 *     if int8 & 0x80 != 0:             # <<<<<<<<<<<<<<
 *         negative = 1
 *         # Need to cast again due to integer promotion
 */
    goto __pyx_L3;
  }

  /* "s3ql/deltadump.pyx":627
 *         int8 = <uint8_t> (int8 & (~ 0x80))
 *     else:
 *         negative = 0             # <<<<<<<<<<<<<<
 * 
 *     if int8 == INT8:
 */
  /*else*/ {
    __pyx_v_negative = 0;
  }
  __pyx_L3:;

  /* "s3ql/deltadump.pyx":629
 *         negative = 0
 * 
 *     if int8 == INT8:             # <<<<<<<<<<<<<<
 *         len_ = 1
 *     elif int8 == INT16:
 */
  __pyx_t_2 = ((__pyx_v_int8 == __pyx_v_4s3ql_9deltadump_INT8) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":630
 * 
 *     if int8 == INT8:
 *         len_ = 1             # <<<<<<<<<<<<<<
 *     elif int8 == INT16:
 *         len_ = 2
 */
    __pyx_v_len_ = 1;

    /* "s3ql/deltadump.pyx":629
 *         negative = 0
 * 
 *     if int8 == INT8:             # <<<<<<<<<<<<<<
 *         len_ = 1
 *     elif int8 == INT16:
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":631
 *     if int8 == INT8:
 *         len_ = 1
 *     elif int8 == INT16:             # <<<<<<<<<<<<<<
 *         len_ = 2
 *     elif int8 == INT32:
 */
  __pyx_t_2 = ((__pyx_v_int8 == __pyx_v_4s3ql_9deltadump_INT16) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":632
 *         len_ = 1
 *     elif int8 == INT16:
 *         len_ = 2             # <<<<<<<<<<<<<<
 *     elif int8 == INT32:
 *         len_ = 4
 */
    __pyx_v_len_ = 2;

    /* "s3ql/deltadump.pyx":631
 *     if int8 == INT8:
 *         len_ = 1
 *     elif int8 == INT16:             # <<<<<<<<<<<<<<
 *         len_ = 2
 *     elif int8 == INT32:
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":633
 *     elif int8 == INT16:
 *         len_ = 2
 *     elif int8 == INT32:             # <<<<<<<<<<<<<<
 *         len_ = 4
 *     elif int8 == INT64:
 */
  __pyx_t_2 = ((__pyx_v_int8 == __pyx_v_4s3ql_9deltadump_INT32) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":634
 *         len_ = 2
 *     elif int8 == INT32:
 *         len_ = 4             # <<<<<<<<<<<<<<
 *     elif int8 == INT64:
 *         len_ = 8
 */
    __pyx_v_len_ = 4;

    /* "s3ql/deltadump.pyx":633
 *     elif int8 == INT16:
 *         len_ = 2
 *     elif int8 == INT32:             # <<<<<<<<<<<<<<
 *         len_ = 4
 *     elif int8 == INT64:
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":635
 *     elif int8 == INT32:
 *         len_ = 4
 *     elif int8 == INT64:             # <<<<<<<<<<<<<<
 *         len_ = 8
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_int8 == __pyx_v_4s3ql_9deltadump_INT64) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":636
 *         len_ = 4
 *     elif int8 == INT64:
 *         len_ = 8             # <<<<<<<<<<<<<<
 *     else:
 *         len_ = 0
 */
    __pyx_v_len_ = 8;

    /* "s3ql/deltadump.pyx":635
 *     elif int8 == INT32:
 *         len_ = 4
 *     elif int8 == INT64:             # <<<<<<<<<<<<<<
 *         len_ = 8
 *     else:
 */
    goto __pyx_L4;
  }

  /* "s3ql/deltadump.pyx":638
 *         len_ = 8
 *     else:
 *         len_ = 0             # <<<<<<<<<<<<<<
 *         uint64 = int8
 * 
 */
  /*else*/ {
    __pyx_v_len_ = 0;

    /* "s3ql/deltadump.pyx":639
 *     else:
 *         len_ = 0
 *         uint64 = int8             # <<<<<<<<<<<<<<
 * 
 *     if len_ != 0:
 */
    __pyx_v_uint64 = __pyx_v_int8;
  }
  __pyx_L4:;

  /* "s3ql/deltadump.pyx":641
 *         uint64 = int8
 * 
 *     if len_ != 0:             # <<<<<<<<<<<<<<
 *         uint64 = 0
 *         fread(&uint64, len_, fp)
 */
  __pyx_t_2 = ((__pyx_v_len_ != 0) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":642
 * 
 *     if len_ != 0:
 *         uint64 = 0             # <<<<<<<<<<<<<<
 *         fread(&uint64, len_, fp)
 *         uint64 = le64toh(uint64)
 */
    __pyx_v_uint64 = 0;

    /* "s3ql/deltadump.pyx":643
 *     if len_ != 0:
 *         uint64 = 0
 *         fread(&uint64, len_, fp)             # <<<<<<<<<<<<<<
 *         uint64 = le64toh(uint64)
 * 
 */
    __pyx_t_1 = __pyx_f_4s3ql_9deltadump_fread((&__pyx_v_uint64), __pyx_v_len_, __pyx_v_fp); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 643, __pyx_L1_error)

    /* "s3ql/deltadump.pyx":644
 *         uint64 = 0
 *         fread(&uint64, len_, fp)
 *         uint64 = le64toh(uint64)             # <<<<<<<<<<<<<<
 * 
 *     if negative == 1:
 */
    __pyx_v_uint64 = le64toh(__pyx_v_uint64);

    /* "s3ql/deltadump.pyx":641
 *         uint64 = int8
 * 
 *     if len_ != 0:             # <<<<<<<<<<<<<<
 *         uint64 = 0
 *         fread(&uint64, len_, fp)
 */
  }

  /* "s3ql/deltadump.pyx":646
 *         uint64 = le64toh(uint64)
 * 
 *     if negative == 1:             # <<<<<<<<<<<<<<
 *         out[0] = - < int64_t > uint64
 *     else:
 */
  __pyx_t_2 = ((__pyx_v_negative == 1) != 0);
  if (__pyx_t_2) {

    /* "s3ql/deltadump.pyx":647
 * 
 *     if negative == 1:
 *         out[0] = - < int64_t > uint64             # <<<<<<<<<<<<<<
 *     else:
 *         out[0] = < int64_t > uint64
 */
    (__pyx_v_out[0]) = (-((int64_t)__pyx_v_uint64));

    /* "s3ql/deltadump.pyx":646
 *         uint64 = le64toh(uint64)
 * 
 *     if negative == 1:             # <<<<<<<<<<<<<<
 *         out[0] = - < int64_t > uint64
 *     else:
 */
    goto __pyx_L6;
  }

  /* "s3ql/deltadump.pyx":649
 *         out[0] = - < int64_t > uint64
 *     else:
 *         out[0] = < int64_t > uint64             # <<<<<<<<<<<<<<
 * 
 *     # len <= 8, safe to cast
 */
  /*else*/ {
    (__pyx_v_out[0]) = ((int64_t)__pyx_v_uint64);
  }
  __pyx_L6:;

  /* "s3ql/deltadump.pyx":652
 * 
 *     # len <= 8, safe to cast
 *     return <int> len_ + 1             # <<<<<<<<<<<<<<
 */
  __pyx_r = (((int)__pyx_v_len_) + 1);
  goto __pyx_L0;

  /* "s3ql/deltadump.pyx":609
 *     return <int> len_ + 1
 * 
 * cdef inline int read_integer(int64_t * out, FILE * fp) except -1:             # <<<<<<<<<<<<<<
 *     '''Read integer written using `write_integer` from *fp*
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("s3ql.deltadump.read_integer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *__pyx_freelist_4s3ql_9deltadump___pyx_scope_struct__dump_table[8];
static int __pyx_freecount_4s3ql_9deltadump___pyx_scope_struct__dump_table = 0;

static PyObject *__pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct__dump_table(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct__dump_table > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table)))) {
    o = (PyObject*)__pyx_freelist_4s3ql_9deltadump___pyx_scope_struct__dump_table[--__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct__dump_table];
    memset(o, 0, sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4s3ql_9deltadump___pyx_scope_struct__dump_table(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct__dump_table < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table)))) {
    __pyx_freelist_4s3ql_9deltadump___pyx_scope_struct__dump_table[__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct__dump_table++] = ((struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_4s3ql_9deltadump___pyx_scope_struct__dump_table = {
  PyVarObject_HEAD_INIT(0, 0)
  "s3ql.deltadump.__pyx_scope_struct__dump_table", /*tp_name*/
  sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct__dump_table), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4s3ql_9deltadump___pyx_scope_struct__dump_table, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct__dump_table, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *__pyx_freelist_4s3ql_9deltadump___pyx_scope_struct_1_load_table[8];
static int __pyx_freecount_4s3ql_9deltadump___pyx_scope_struct_1_load_table = 0;

static PyObject *__pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct_1_load_table(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct_1_load_table > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table)))) {
    o = (PyObject*)__pyx_freelist_4s3ql_9deltadump___pyx_scope_struct_1_load_table[--__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct_1_load_table];
    memset(o, 0, sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_4s3ql_9deltadump___pyx_scope_struct_1_load_table(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct_1_load_table < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table)))) {
    __pyx_freelist_4s3ql_9deltadump___pyx_scope_struct_1_load_table[__pyx_freecount_4s3ql_9deltadump___pyx_scope_struct_1_load_table++] = ((struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_type_4s3ql_9deltadump___pyx_scope_struct_1_load_table = {
  PyVarObject_HEAD_INIT(0, 0)
  "s3ql.deltadump.__pyx_scope_struct_1_load_table", /*tp_name*/
  sizeof(struct __pyx_obj_4s3ql_9deltadump___pyx_scope_struct_1_load_table), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_4s3ql_9deltadump___pyx_scope_struct_1_load_table, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_4s3ql_9deltadump___pyx_scope_struct_1_load_table, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "deltadump",
    __pyx_k_deltadump_pyx_this_file_is_part, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_b_BEGIN_TRANSACTION, __pyx_k_BEGIN_TRANSACTION, sizeof(__pyx_k_BEGIN_TRANSACTION), 0, 0, 0, 0},
  {&__pyx_n_s_BLOB, __pyx_k_BLOB, sizeof(__pyx_k_BLOB), 0, 0, 1, 1},
  {&__pyx_kp_u_BLOB_too_large_to_read_d_vs_d, __pyx_k_BLOB_too_large_to_read_d_vs_d, sizeof(__pyx_k_BLOB_too_large_to_read_d_vs_d), 0, 1, 0, 0},
  {&__pyx_kp_b_COMMIT_TRANSACTION, __pyx_k_COMMIT_TRANSACTION, sizeof(__pyx_k_COMMIT_TRANSACTION), 0, 0, 0, 0},
  {&__pyx_kp_u_Can_not_dump_BLOB_of_size_d_max, __pyx_k_Can_not_dump_BLOB_of_size_d_max, sizeof(__pyx_k_Can_not_dump_BLOB_of_size_d_max), 0, 1, 0, 0},
  {&__pyx_kp_u_Can_t_access_in_memory_databases, __pyx_k_Can_t_access_in_memory_databases, sizeof(__pyx_k_Can_t_access_in_memory_databases), 0, 1, 0, 0},
  {&__pyx_kp_u_Can_t_dump_NULL_values, __pyx_k_Can_t_dump_NULL_values, sizeof(__pyx_k_Can_t_dump_NULL_values), 0, 1, 0, 0},
  {&__pyx_kp_u_Corrupted_input, __pyx_k_Corrupted_input, sizeof(__pyx_k_Corrupted_input), 0, 1, 0, 0},
  {&__pyx_n_s_ExitStack, __pyx_k_ExitStack, sizeof(__pyx_k_ExitStack), 0, 0, 1, 1},
  {&__pyx_kp_u_INSERT_INTO_s_s_VALUES_s, __pyx_k_INSERT_INTO_s_s_VALUES_s, sizeof(__pyx_k_INSERT_INTO_s_s_VALUES_s), 0, 1, 0, 0},
  {&__pyx_n_s_INTEGER, __pyx_k_INTEGER, sizeof(__pyx_k_INTEGER), 0, 0, 1, 1},
  {&__pyx_n_s_IOError, __pyx_k_IOError, sizeof(__pyx_k_IOError), 0, 0, 1, 1},
  {&__pyx_kp_u_Invalid_type_for_column_d, __pyx_k_Invalid_type_for_column_d, sizeof(__pyx_k_Invalid_type_for_column_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Length_d_d_in_column_d, __pyx_k_Length_d_d_in_column_d, sizeof(__pyx_k_Length_d_d_in_column_d), 0, 1, 0, 0},
  {&__pyx_n_s_MAX_BLOB_SIZE, __pyx_k_MAX_BLOB_SIZE, sizeof(__pyx_k_MAX_BLOB_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_kp_u_Opening_connection_to_s, __pyx_k_Opening_connection_to_s, sizeof(__pyx_k_Opening_connection_to_s), 0, 1, 0, 0},
  {&__pyx_kp_u_PRAGMA_s, __pyx_k_PRAGMA_s, sizeof(__pyx_k_PRAGMA_s), 0, 1, 0, 0},
  {&__pyx_kp_u_PRAGMA_s_s, __pyx_k_PRAGMA_s_s, sizeof(__pyx_k_PRAGMA_s_s), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_kp_u_SELECT_COUNT_rowid_FROM_s, __pyx_k_SELECT_COUNT_rowid_FROM_s, sizeof(__pyx_k_SELECT_COUNT_rowid_FROM_s), 0, 1, 0, 0},
  {&__pyx_kp_u_SELECT_s_FROM_s_ORDER_BY_s, __pyx_k_SELECT_s_FROM_s_ORDER_BY_s, sizeof(__pyx_k_SELECT_s_FROM_s_ORDER_BY_s), 0, 1, 0, 0},
  {&__pyx_kp_u_SQLite_code_used_by_APSW_was_com, __pyx_k_SQLite_code_used_by_APSW_was_com, sizeof(__pyx_k_SQLite_code_used_by_APSW_was_com), 0, 1, 0, 0},
  {&__pyx_kp_u_SQLite_version_mismatch_between, __pyx_k_SQLite_version_mismatch_between, sizeof(__pyx_k_SQLite_version_mismatch_between), 0, 1, 0, 0},
  {&__pyx_n_s_TIME, __pyx_k_TIME, sizeof(__pyx_k_TIME), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
  {&__pyx_n_s_apsw, __pyx_k_apsw, sizeof(__pyx_k_apsw), 0, 0, 1, 1},
  {&__pyx_kp_u_apsw_sqlite_compile_options_s_s3, __pyx_k_apsw_sqlite_compile_options_s_s3, sizeof(__pyx_k_apsw_sqlite_compile_options_s_s3), 0, 1, 0, 0},
  {&__pyx_n_s_apsw_sqlite_options, __pyx_k_apsw_sqlite_options, sizeof(__pyx_k_apsw_sqlite_options), 0, 0, 1, 1},
  {&__pyx_n_s_apsw_sqlite_version, __pyx_k_apsw_sqlite_version, sizeof(__pyx_k_apsw_sqlite_version), 0, 0, 1, 1},
  {&__pyx_kp_u_apsw_sqlite_version_s_s3ql_sqlit, __pyx_k_apsw_sqlite_version_s_s3ql_sqlit, sizeof(__pyx_k_apsw_sqlite_version_s_s3ql_sqlit), 0, 1, 0, 0},
  {&__pyx_n_s_begin_stmt, __pyx_k_begin_stmt, sizeof(__pyx_k_begin_stmt), 0, 0, 1, 1},
  {&__pyx_n_s_buf, __pyx_k_buf, sizeof(__pyx_k_buf), 0, 0, 1, 1},
  {&__pyx_n_s_callback, __pyx_k_callback, sizeof(__pyx_k_callback), 0, 0, 1, 1},
  {&__pyx_n_s_check_sqlite, __pyx_k_check_sqlite, sizeof(__pyx_k_check_sqlite), 0, 0, 1, 1},
  {&__pyx_n_s_cm, __pyx_k_cm, sizeof(__pyx_k_cm), 0, 0, 1, 1},
  {&__pyx_n_s_cmd, __pyx_k_cmd, sizeof(__pyx_k_cmd), 0, 0, 1, 1},
  {&__pyx_n_s_col_args, __pyx_k_col_args, sizeof(__pyx_k_col_args), 0, 0, 1, 1},
  {&__pyx_n_s_col_count, __pyx_k_col_count, sizeof(__pyx_k_col_count), 0, 0, 1, 1},
  {&__pyx_n_s_col_names, __pyx_k_col_names, sizeof(__pyx_k_col_names), 0, 0, 1, 1},
  {&__pyx_n_s_col_types, __pyx_k_col_types, sizeof(__pyx_k_col_types), 0, 0, 1, 1},
  {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
  {&__pyx_n_s_commit_stmt, __pyx_k_commit_stmt, sizeof(__pyx_k_commit_stmt), 0, 0, 1, 1},
  {&__pyx_n_s_compile_options, __pyx_k_compile_options, sizeof(__pyx_k_compile_options), 0, 0, 1, 1},
  {&__pyx_n_s_contextlib, __pyx_k_contextlib, sizeof(__pyx_k_contextlib), 0, 0, 1, 1},
  {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
  {&__pyx_n_s_db, __pyx_k_db, sizeof(__pyx_k_db), 0, 0, 1, 1},
  {&__pyx_n_s_dbfile_b, __pyx_k_dbfile_b, sizeof(__pyx_k_dbfile_b), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_dump_table, __pyx_k_dump_table, sizeof(__pyx_k_dump_table), 0, 0, 1, 1},
  {&__pyx_n_s_dump_table_locals_lambda, __pyx_k_dump_table_locals_lambda, sizeof(__pyx_k_dump_table_locals_lambda), 0, 0, 1, 1},
  {&__pyx_kp_u_dump_table_s_writing_d_rows, __pyx_k_dump_table_s_writing_d_rows, sizeof(__pyx_k_dump_table_s_writing_d_rows), 0, 1, 0, 0},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_exceptionfor, __pyx_k_exceptionfor, sizeof(__pyx_k_exceptionfor), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_fh, __pyx_k_fh, sizeof(__pyx_k_fh), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_fileno, __pyx_k_fileno, sizeof(__pyx_k_fileno), 0, 0, 1, 1},
  {&__pyx_n_u_foreign_keys, __pyx_k_foreign_keys, sizeof(__pyx_k_foreign_keys), 0, 1, 0, 1},
  {&__pyx_n_s_fp, __pyx_k_fp, sizeof(__pyx_k_fp), 0, 0, 1, 1},
  {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
  {&__pyx_n_s_get_val, __pyx_k_get_val, sizeof(__pyx_k_get_val), 0, 0, 1, 1},
  {&__pyx_n_s_getfilesystemencoding, __pyx_k_getfilesystemencoding, sizeof(__pyx_k_getfilesystemencoding), 0, 0, 1, 1},
  {&__pyx_kp_s_home_nikratio_in_progress_s3ql, __pyx_k_home_nikratio_in_progress_s3ql, sizeof(__pyx_k_home_nikratio_in_progress_s3ql), 0, 0, 1, 0},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 0, 1, 1},
  {&__pyx_n_s_int64_prev, __pyx_k_int64_prev, sizeof(__pyx_k_int64_prev), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_load_table, __pyx_k_load_table, sizeof(__pyx_k_load_table), 0, 0, 1, 1},
  {&__pyx_n_s_load_table_locals_lambda, __pyx_k_load_table_locals_lambda, sizeof(__pyx_k_load_table_locals_lambda), 0, 0, 1, 1},
  {&__pyx_kp_u_load_table_s_reading_d_rows, __pyx_k_load_table_s_reading_d_rows, sizeof(__pyx_k_load_table_s_reading_d_rows), 0, 1, 0, 0},
  {&__pyx_n_s_log, __pyx_k_log, sizeof(__pyx_k_log), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_kp_u_memory, __pyx_k_memory, sizeof(__pyx_k_memory), 0, 1, 0, 0},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_pragma, __pyx_k_pragma, sizeof(__pyx_k_pragma), 0, 0, 1, 1},
  {&__pyx_n_s_query, __pyx_k_query, sizeof(__pyx_k_query), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rc, __pyx_k_rc, sizeof(__pyx_k_rc), 0, 0, 1, 1},
  {&__pyx_n_s_row_count, __pyx_k_row_count, sizeof(__pyx_k_row_count), 0, 0, 1, 1},
  {&__pyx_n_s_s3ql_deltadump, __pyx_k_s3ql_deltadump, sizeof(__pyx_k_s3ql_deltadump), 0, 0, 1, 1},
  {&__pyx_n_s_s3ql_sqlite_options, __pyx_k_s3ql_sqlite_options, sizeof(__pyx_k_s3ql_sqlite_options), 0, 0, 1, 1},
  {&__pyx_n_s_s3ql_sqlite_version, __pyx_k_s3ql_sqlite_version, sizeof(__pyx_k_s3ql_sqlite_version), 0, 0, 1, 1},
  {&__pyx_n_s_sqlite3_db, __pyx_k_sqlite3_db, sizeof(__pyx_k_sqlite3_db), 0, 0, 1, 1},
  {&__pyx_n_s_sqlitelibversion, __pyx_k_sqlitelibversion, sizeof(__pyx_k_sqlitelibversion), 0, 0, 1, 1},
  {&__pyx_n_s_stmt, __pyx_k_stmt, sizeof(__pyx_k_stmt), 0, 0, 1, 1},
  {&__pyx_n_u_surrogateescape, __pyx_k_surrogateescape, sizeof(__pyx_k_surrogateescape), 0, 1, 0, 1},
  {&__pyx_n_u_synchronous, __pyx_k_synchronous, sizeof(__pyx_k_synchronous), 0, 1, 0, 1},
  {&__pyx_n_s_sys, __pyx_k_sys, sizeof(__pyx_k_sys), 0, 0, 1, 1},
  {&__pyx_n_s_table, __pyx_k_table, sizeof(__pyx_k_table), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tmp, __pyx_k_tmp, sizeof(__pyx_k_tmp), 0, 0, 1, 1},
  {&__pyx_n_s_trx_rows, __pyx_k_trx_rows, sizeof(__pyx_k_trx_rows), 0, 0, 1, 1},
  {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 136, __pyx_L1_error)
  __pyx_builtin_IOError = __Pyx_GetBuiltinName(__pyx_n_s_IOError); if (!__pyx_builtin_IOError) __PYX_ERR(0, 116, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 206, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 208, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 258, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "s3ql/deltadump.pyx":263
 *     apsw_sqlite_options = set(apsw.compile_options)
 *     s3ql_sqlite_options = set()
 *     for idx in itertools.count(0):             # <<<<<<<<<<<<<<
 *         buf = sqlite3_compileoption_get(idx)
 *         if buf is NULL:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "s3ql/deltadump.pyx":319
 * 
 *     if db.file == ':memory:':
 *         raise ValueError("Can't access in-memory databases")             # <<<<<<<<<<<<<<
 * 
 *     with ExitStack() as cm:
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_Can_t_access_in_memory_databases); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "s3ql/deltadump.pyx":368
 *             for i in range(col_count):
 *                 if sqlite3_column_type(stmt, i) is SQLITE_NULL:
 *                     raise ValueError("Can't dump NULL values")             # <<<<<<<<<<<<<<
 * 
 *                 if col_types[i] == _INTEGER:
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_u_Can_t_dump_NULL_values); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "s3ql/deltadump.pyx":321
 *         raise ValueError("Can't access in-memory databases")
 * 
 *     with ExitStack() as cm:             # <<<<<<<<<<<<<<
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 */
  __pyx_tuple__6 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "s3ql/deltadump.pyx":432
 * 
 *     if db.file == ':memory:':
 *         raise ValueError("Can't access in-memory databases")             # <<<<<<<<<<<<<<
 * 
 *     with ExitStack() as cm:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Can_t_access_in_memory_databases); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "s3ql/deltadump.pyx":447
 * 
 *         # Copy settings
 *         for pragma in ('synchronous', 'foreign_keys'):             # <<<<<<<<<<<<<<
 *             val = db.get_val('PRAGMA %s' % pragma)
 *             cmd = ('PRAGMA %s = %s' % (pragma, val)).encode('utf-8')
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_n_u_synchronous, __pyx_n_u_foreign_keys); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "s3ql/deltadump.pyx":532
 *                         tmp = int64_prev[j] + int64
 *                         if tmp < 0 or tmp > INT_MAX:
 *                             raise RuntimeError('Corrupted input')             # <<<<<<<<<<<<<<
 *                         len_ = <int> tmp
 *                         int64_prev[j] = tmp
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_Corrupted_input); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "s3ql/deltadump.pyx":434
 *         raise ValueError("Can't access in-memory databases")
 * 
 *     with ExitStack() as cm:             # <<<<<<<<<<<<<<
 *         # Get SQLite connection
 *         log.debug('Opening connection to %s', db.file)
 */
  __pyx_tuple__11 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "s3ql/deltadump.pyx":237
 *     return fp
 * 
 * def check_sqlite():             # <<<<<<<<<<<<<<
 *     '''Check if deltadump and apsw module use compatible SQLite code.
 * 
 */
  __pyx_tuple__12 = PyTuple_Pack(6, __pyx_n_s_buf, __pyx_n_s_apsw_sqlite_version, __pyx_n_s_s3ql_sqlite_version, __pyx_n_s_apsw_sqlite_options, __pyx_n_s_s3ql_sqlite_options, __pyx_n_s_idx); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(0, 0, 6, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__12, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_nikratio_in_progress_s3ql, __pyx_n_s_check_sqlite, 237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 237, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":280
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 * def dump_table(table, order, columns, db, fh):             # <<<<<<<<<<<<<<
 *     '''Dump *columns* of *table* into *fh*
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(23, __pyx_n_s_table, __pyx_n_s_order, __pyx_n_s_columns, __pyx_n_s_db, __pyx_n_s_fh, __pyx_n_s_sqlite3_db, __pyx_n_s_stmt, __pyx_n_s_col_types, __pyx_n_s_col_args, __pyx_n_s_col_count, __pyx_n_s_rc, __pyx_n_s_i, __pyx_n_s_len, __pyx_n_s_int64_prev, __pyx_n_s_int64, __pyx_n_s_tmp, __pyx_n_s_fp, __pyx_n_s_buf, __pyx_n_s_row_count, __pyx_n_s_cm, __pyx_n_s_dbfile_b, __pyx_n_s_col_names, __pyx_n_s_query); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(5, 0, 23, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__14, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_nikratio_in_progress_s3ql, __pyx_n_s_dump_table, 280, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 280, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":401
 *                         fwrite(buf, len_, fp)
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):             # <<<<<<<<<<<<<<
 *     '''Load *columns* of *table* from *fh*
 * 
 */
  __pyx_tuple__16 = PyTuple_Pack(29, __pyx_n_s_table, __pyx_n_s_columns, __pyx_n_s_db, __pyx_n_s_fh, __pyx_n_s_trx_rows, __pyx_n_s_sqlite3_db, __pyx_n_s_stmt, __pyx_n_s_begin_stmt, __pyx_n_s_commit_stmt, __pyx_n_s_col_types, __pyx_n_s_col_args, __pyx_n_s_col_count, __pyx_n_s_rc, __pyx_n_s_len, __pyx_n_s_i, __pyx_n_s_j, __pyx_n_s_int64_prev, __pyx_n_s_fp, __pyx_n_s_buf, __pyx_n_s_row_count, __pyx_n_s_int64, __pyx_n_s_tmp, __pyx_n_s_cm, __pyx_n_s_dbfile_b, __pyx_n_s_pragma, __pyx_n_s_val, __pyx_n_s_cmd, __pyx_n_s_col_names, __pyx_n_s_query); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(5, 0, 29, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__16, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_nikratio_in_progress_s3ql, __pyx_n_s_load_table, 401, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4096 = PyInt_FromLong(4096); if (unlikely(!__pyx_int_4096)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5000 = PyInt_FromLong(5000); if (unlikely(!__pyx_int_5000)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initdeltadump(void); /*proto*/
PyMODINIT_FUNC initdeltadump(void)
#else
PyMODINIT_FUNC PyInit_deltadump(void); /*proto*/
PyMODINIT_FUNC PyInit_deltadump(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_deltadump(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("deltadump", __pyx_methods, __pyx_k_deltadump_pyx_this_file_is_part, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_s3ql__deltadump) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "s3ql.deltadump")) {
      if (unlikely(PyDict_SetItemString(modules, "s3ql.deltadump", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  if (PyType_Ready(&__pyx_type_4s3ql_9deltadump___pyx_scope_struct__dump_table) < 0) __PYX_ERR(0, 280, __pyx_L1_error)
  __pyx_type_4s3ql_9deltadump___pyx_scope_struct__dump_table.tp_print = 0;
  __pyx_ptype_4s3ql_9deltadump___pyx_scope_struct__dump_table = &__pyx_type_4s3ql_9deltadump___pyx_scope_struct__dump_table;
  if (PyType_Ready(&__pyx_type_4s3ql_9deltadump___pyx_scope_struct_1_load_table) < 0) __PYX_ERR(0, 401, __pyx_L1_error)
  __pyx_type_4s3ql_9deltadump___pyx_scope_struct_1_load_table.tp_print = 0;
  __pyx_ptype_4s3ql_9deltadump___pyx_scope_struct_1_load_table = &__pyx_type_4s3ql_9deltadump___pyx_scope_struct_1_load_table;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(1, 9, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "s3ql/deltadump.pyx":79
 *         SQLITE_OPEN_READONLY
 * 
 * from contextlib import ExitStack             # <<<<<<<<<<<<<<
 * import apsw
 * import os
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_ExitStack);
  __Pyx_GIVEREF(__pyx_n_s_ExitStack);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ExitStack);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_contextlib, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ExitStack); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ExitStack, __pyx_t_1) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "s3ql/deltadump.pyx":80
 * 
 * from contextlib import ExitStack
 * import apsw             # <<<<<<<<<<<<<<
 * import os
 * from .logging import logging # Ensure use of custom logger class
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_apsw, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_apsw, __pyx_t_2) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "s3ql/deltadump.pyx":81
 * from contextlib import ExitStack
 * import apsw
 * import os             # <<<<<<<<<<<<<<
 * from .logging import logging # Ensure use of custom logger class
 * import itertools
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "s3ql/deltadump.pyx":82
 * import apsw
 * import os
 * from .logging import logging # Ensure use of custom logger class             # <<<<<<<<<<<<<<
 * import itertools
 * import sys
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_logging);
  __Pyx_GIVEREF(__pyx_n_s_logging);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_logging);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_logging, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_2) < 0) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":83
 * import os
 * from .logging import logging # Ensure use of custom logger class
 * import itertools             # <<<<<<<<<<<<<<
 * import sys
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_itertools, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_itertools, __pyx_t_1) < 0) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":84
 * from .logging import logging # Ensure use of custom logger class
 * import itertools
 * import sys             # <<<<<<<<<<<<<<
 * 
 * log = logging.getLogger(__name__)
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_sys, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sys, __pyx_t_1) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":86
 * import sys
 * 
 * log = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 * 
 * # Column types
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_logging); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_log, __pyx_t_1) < 0) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":89
 * 
 * # Column types
 * cdef int _INTEGER = 1             # <<<<<<<<<<<<<<
 * cdef int _BLOB = 2
 * cdef int _TIME = 3
 */
  __pyx_v_4s3ql_9deltadump__INTEGER = 1;

  /* "s3ql/deltadump.pyx":90
 * # Column types
 * cdef int _INTEGER = 1
 * cdef int _BLOB = 2             # <<<<<<<<<<<<<<
 * cdef int _TIME = 3
 * 
 */
  __pyx_v_4s3ql_9deltadump__BLOB = 2;

  /* "s3ql/deltadump.pyx":91
 * cdef int _INTEGER = 1
 * cdef int _BLOB = 2
 * cdef int _TIME = 3             # <<<<<<<<<<<<<<
 * 
 * # Make column types available as Python objects
 */
  __pyx_v_4s3ql_9deltadump__TIME = 3;

  /* "s3ql/deltadump.pyx":94
 * 
 * # Make column types available as Python objects
 * INTEGER = _INTEGER             # <<<<<<<<<<<<<<
 * BLOB = _BLOB
 * TIME = _TIME
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_4s3ql_9deltadump__INTEGER); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INTEGER, __pyx_t_1) < 0) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":95
 * # Make column types available as Python objects
 * INTEGER = _INTEGER
 * BLOB = _BLOB             # <<<<<<<<<<<<<<
 * TIME = _TIME
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_4s3ql_9deltadump__BLOB); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BLOB, __pyx_t_1) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":96
 * INTEGER = _INTEGER
 * BLOB = _BLOB
 * TIME = _TIME             # <<<<<<<<<<<<<<
 * 
 * # Integer length codes
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_4s3ql_9deltadump__TIME); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TIME, __pyx_t_1) < 0) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":99
 * 
 * # Integer length codes
 * cdef uint8_t INT8 = 127             # <<<<<<<<<<<<<<
 * cdef uint8_t INT16 = 126
 * cdef uint8_t INT32 = 125
 */
  __pyx_v_4s3ql_9deltadump_INT8 = 0x7F;

  /* "s3ql/deltadump.pyx":100
 * # Integer length codes
 * cdef uint8_t INT8 = 127
 * cdef uint8_t INT16 = 126             # <<<<<<<<<<<<<<
 * cdef uint8_t INT32 = 125
 * cdef uint8_t INT64 = 124
 */
  __pyx_v_4s3ql_9deltadump_INT16 = 0x7E;

  /* "s3ql/deltadump.pyx":101
 * cdef uint8_t INT8 = 127
 * cdef uint8_t INT16 = 126
 * cdef uint8_t INT32 = 125             # <<<<<<<<<<<<<<
 * cdef uint8_t INT64 = 124
 * 
 */
  __pyx_v_4s3ql_9deltadump_INT32 = 0x7D;

  /* "s3ql/deltadump.pyx":102
 * cdef uint8_t INT16 = 126
 * cdef uint8_t INT32 = 125
 * cdef uint8_t INT64 = 124             # <<<<<<<<<<<<<<
 * 
 * # Maximum size of BLOBs
 */
  __pyx_v_4s3ql_9deltadump_INT64 = 0x7C;

  /* "s3ql/deltadump.pyx":105
 * 
 * # Maximum size of BLOBs
 * MAX_BLOB_SIZE = 4096             # <<<<<<<<<<<<<<
 * 
 * # Scale factor from time floats to integers. 1e9 would give nanosecond
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAX_BLOB_SIZE, __pyx_int_4096) < 0) __PYX_ERR(0, 105, __pyx_L1_error)

  /* "s3ql/deltadump.pyx":110
 * # resolution but introduces rounding errors, so we use 1 << 30 (which is
 * # ~1.074e9, i.e. we get a little more precision than nanoseconds).
 * cdef double time_scale = 1 << 30             # <<<<<<<<<<<<<<
 * 
 * cdef inline int fwrite(const_void * buf, size_t len_, FILE * fp) except -1:
 */
  __pyx_v_4s3ql_9deltadump_time_scale = 1073741824.0;

  /* "s3ql/deltadump.pyx":136
 *     return None
 * 
 * cdef int raise_from_errno(err_class=OSError) except -1:             # <<<<<<<<<<<<<<
 *     '''Raise OSError for current errno value'''
 * 
 */
  __Pyx_INCREF(__pyx_builtin_OSError);
  __pyx_k_ = __pyx_builtin_OSError;
  __Pyx_GIVEREF(__pyx_builtin_OSError);

  /* "s3ql/deltadump.pyx":237
 *     return fp
 * 
 * def check_sqlite():             # <<<<<<<<<<<<<<
 *     '''Check if deltadump and apsw module use compatible SQLite code.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_1check_sqlite, NULL, __pyx_n_s_s3ql_deltadump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_check_sqlite, __pyx_t_1) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":280
 *                            s3ql_sqlite_options - apsw_sqlite_options))
 * 
 * def dump_table(table, order, columns, db, fh):             # <<<<<<<<<<<<<<
 *     '''Dump *columns* of *table* into *fh*
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_3dump_table, NULL, __pyx_n_s_s3ql_deltadump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dump_table, __pyx_t_1) < 0) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":401
 *                         fwrite(buf, len_, fp)
 * 
 * def load_table(table, columns, db, fh, trx_rows=5000):             # <<<<<<<<<<<<<<
 *     '''Load *columns* of *table* from *fh*
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_4s3ql_9deltadump_5load_table, NULL, __pyx_n_s_s3ql_deltadump); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_load_table, __pyx_t_1) < 0) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "s3ql/deltadump.pyx":1
 * '''             # <<<<<<<<<<<<<<
 * deltadump.pyx - this file is part of S3QL (http://s3ql.googlecode.com)
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init s3ql.deltadump", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init s3ql.deltadump");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_COMPILING_IN_CPYTHON
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_COMPILING_IN_CPYTHON
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                #if PyLong_SHIFT < 30 && PyLong_SHIFT != 15
                default: return PyLong_Type.tp_richcompare(op1, op2, Py_EQ);
                #else
                default: Py_RETURN_FALSE;
                #endif
            }
        }
            if (a == b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            if ((double)a == (double)b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    return PyObject_RichCompare(op1, op2, Py_EQ);
}
#endif

static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
    }
}

static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) -digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int64_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int64_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 2 * PyLong_SHIFT) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 3 * PyLong_SHIFT) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) >= 4 * PyLong_SHIFT) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int64_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int64_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int64_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int64_t, sdigit, (sdigit) -digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(int64_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int64_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int64_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int64_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int64_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
            } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int64_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int64_t) -1;
        }
    } else {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) -digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) -digits[0])
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

#ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
